/*******************************************************************************\

	main.c in Sequential Minimal Optimization ver2.0
		
	entry function.
		
	Chu Wei Copyright(C) National Univeristy of Singapore
	Create on Jan. 16 2000 at Control Lab of Mechanical Engineering 
	Update on Aug. 23 2001 

\*******************************************************************************/

#include <stdio.h>
#include <stdlib.h>
#include <malloc.h>
#include <string.h>
#include <limits.h>
#include <math.h>
#include <time.h>
#include "smo.h"
#define VERSION (0)

int main( int argc, char * argv[])
{
	def_Settings * defsetting = NULL ;
	kcv_Settings * kcvsetting ;
	//Data_Node * node ;
	char buf[LENGTH] ;
	char filename[1024] ;
	unsigned int sz = 0;
	unsigned int index = 0 ;
	double parameter = 0 ;
	printf("\nK-fold generator v2.%d \n--- Chu Wei Copyright(C) 2003-2004\n\n", VERSION) ;
	if ( 1 == argc || NULL == (defsetting = Create_def_Settings(argv[--argc])) )
	{
		// display help	
		printf("\nUsage:  svor [-v] [...] [-R r] file \n\n") ;
		printf("  file   specifies the file containing training samples.\n") ;
		printf("  -v     activates the verbose mode to display message.\n") ;
		printf("  -F  k  k-fold is used in cross validation (default 10 folds).\n") ;
		printf("  -S  s  specify the seed of random number generator at s.\n") ;
		printf("\n") ;
		if (NULL !=defsetting)
			Clear_def_Settings( defsetting ) ;
		return 0;
	}
	else
	{
		if (argc>1)
			printf("Options:\n") ;
		do
		{
			strcpy(buf, argv[--argc]) ;
			sz = strlen(buf) ;
			//printf ("%s  %d\n", buf, sz) ;
			if ( '-' == buf[0] )
			{				
				for (index = 1 ; index < sz ; index++)
				{
					switch (buf[index])
					{
					case 'v' :
						printf("  - Verbose mode in display.\n") ;
						defsetting->smo_display = TRUE ;
						break ;	
					case 'a' :
						printf("  - kernels with Ard parameters.\n") ;
						defsetting->ardon = TRUE ;	
						break ;
					case 'o' :
						printf("  - normalize the Outputs in training data.\n") ;
						defsetting->normalized_output = TRUE ;	
						defsetting->pairs.normalized_output = TRUE ;
						break ;
					case 'i' :
						printf("  - normalize the Inputs in training data.\n") ;
						defsetting->normalized_input = TRUE ;	
						defsetting->pairs.normalized_input = TRUE ;
						break ;
					case 'S' :
						if (parameter>0)
						{
							printf("  - specify %.0f as the seed of random number.\n", parameter) ;
							defsetting->seeds = (unsigned int)parameter ;
						}
						break ;
					case 'F' :
						if (parameter>1)
						{
							printf("  - %.0f-fold cross validation.\n", parameter) ;
							defsetting->kfold = (unsigned int)parameter ;
						}
						break ;
					case 'Z' :
						if (parameter>=1)
						{
							printf("  - Zoom in scale %.0f.\n", parameter) ;
							defsetting->loops = (unsigned int)parameter ;
						}
						break ;
					case 'R' :
						if (parameter>1)
						{
							printf("  - Rehearsal %.0f times.\n", parameter) ;
							defsetting->repeat = (unsigned int)parameter ;
						}
						break ;
					case 'L' :
						printf("  - choose Linear kernel.\n") ;
						defsetting->kernel = LINEAR ;						
						break ;
					case 'B' :
						printf("  - set regularization factors balanced.\n") ;
						defsetting->smo_balance = TRUE ;
						break ;
					case 'T' :
						if (parameter>0)
						{
							printf("  - set Tol as %.6f.\n", parameter) ;
							defsetting->tol = parameter ;
						}
						break ;
					case 'C' :
						if (parameter > 0)
						{ 
							if (index + 1 == sz && parameter != DEF_COARSESTEP )
							{
								printf("  - C is invalid.\n") ;
								parameter = 0;
							}
							else if (index+1==sz-1)
							{
								index+=1;
								switch (buf[index])
								{									
								case 's' :
									defsetting->def_lnC_step = parameter ;
									printf("  - initialize lnC_step at the value %f.\n", parameter) ;
									parameter = 0 ;
									break ;
								case 'c' :
									defsetting->def_lnC_start = log10(parameter) ;
									printf("  - C start at %f.\n", parameter) ;
									parameter = 0 ;
									break ;
								case 'e' :
									defsetting->def_lnC_end = log10(parameter) ;
									printf("  - C end at %f.\n", parameter) ;
									parameter = 0 ;
									break ;
								default :
									printf("  - C%c is invalid.\n", buf[index]) ;
									break ;
								}
							}							
						}
						break ;						
					case 'K' :
						if (parameter > 0)
						{ 
							if (index + 1 == sz && parameter != DEF_COARSESTEP )
							{
								printf("  - K is invalid.\n") ;
								parameter = 0;
							}
							else if (index+1==sz-1)
							{
								index+=1;
								switch (buf[index])
								{									
								case 's' :
									defsetting->def_lnK_step = parameter ;
									printf("  - initialize lnK_step at the value %f.\n", parameter) ;
									parameter = 0 ;
									break ;
								case 'c' :
									defsetting->def_lnK_start = log10(parameter) ;
									printf("  - K start at %f.\n", parameter) ;
									parameter = 0 ;
									break ;
								case 'e' :
									defsetting->def_lnK_end = log10(parameter) ;
									printf("  - K end at %f.\n", parameter) ;
									parameter = 0 ;
									break ;
								default :
									printf("  - K%c is invalid.\n", buf[index]) ;
									break ;
								}
							}							
						}
						break ;
					case 'P' :						
						if (parameter >= 1)
						{ 
							defsetting->kernel = POLYNOMIAL ;
							defsetting->p = (unsigned int) parameter ;
							printf("  - choose Polynomial kernel with order %d.\n", defsetting->p) ;
							parameter = 0 ;
							defsetting->def_lnK_start = 0 ;
							defsetting->def_lnK_end = 0 ;
						}					
						break ;	
					default :
						if ('-' != buf[index])
							printf("  -%c is invalid.\n", buf[index]) ;
						break ;
					}
				}
			}
			else
				parameter = atof(buf) ;
		}
		while ( argc > 1 ) ;
		printf("\n") ;
	}

	sprintf(filename,"%s",defsetting->inputfile) ;
	if ( TRUE == Update_def_Settings(defsetting) ) 
	{
		sz = defsetting->loops ;
		while (sz > 0)
		{
			// coarse search
			kcvsetting = Create_Kcv ( defsetting ) ;
			while (kcvsetting->index<defsetting->repeat)
			{
				Rehearsal_Kcv ( kcvsetting, defsetting ) ;
				Init_Kcv ( kcvsetting, defsetting ) ;
				kcvsetting->index += 1 ; 
			}			
			// update defsetting 
			Clear_Kcv ( kcvsetting ) ;
			sz -= 1 ;
		}
	}
	// free memory then exit
	Clear_def_Settings( defsetting ) ;	
	return 0;
}
//end of main.c 
