#include <stdio.h>
#include <stdlib.h>
#include <string.h>
#include <time.h>
#include <math.h>
#include "ordinal.h"
//#include "bit_matrix.h"

extern double lerrf(double) ;
extern double fromb(double mean, double sig, double eps) ;

double Ordinal_Prediction (Data_List * testlist, Ordinal_Settings * settings) 
{
	Data_List * trainlist ;
	Data_Node * node ;
	unsigned int i, k ;
	double * kcvx ;
	double * prob ;
	double covx ;
	double z1, z2 ;
	unsigned int index = 0 ;
	double phi1, phi2 ;
	double guess ;
	int error = 0, abserr = 0 ;
	double logprob = 0 ;

	if (NULL == settings || NULL == testlist)
		return 1 ;
	
	trainlist = settings->pairs ;
	
	// compute predictions
	node = testlist->front ;
	while (NULL != node)
	{
		// for each new point
		// transform input point
		node->pred_func = 0 ;
		// cache kcx 
		kcvx = (double*)malloc(trainlist->count*sizeof(double)) ;
		for (i=0;i<trainlist->count;i++)
		{
			kcvx[i] = Calc_Covfun((settings->alpha+i)->pair->point,node->point,settings) ;
			node->pred_func += (settings->alpha+i)->pair->weight*kcvx[i];
		}
		covx = Calc_Covfun(node->point,node->point,settings) ; ;
		// compute sigmax
		for (i=0;i<trainlist->count;i++)
		{
			for (k=0;k<i;k++)
				covx -= 2.0*kcvx[i]*kcvx[k]*settings->invcov[i*trainlist->count+k] ;
			covx -= kcvx[i]*kcvx[i]*settings->invcov[i*trainlist->count+i] ;
		}
		if (covx>0)
			node->pred_var = covx ;
		else
			printf("Warning: %u sample with a negative variance.",index+1) ;
		free(kcvx) ;
		node=node->next ;
		index += 1 ;
	}

	// compute predictions
	prob = (double*) malloc(trainlist->classes*sizeof(double)) ;
	node = testlist->front ;	
	while (NULL != node)
	{
		// compute predictive probability
		guess = 0 ;
		for (i=1;i<=trainlist->classes;i++)
		{			
			phi2 = 0 ;
			phi1 = 1 ;
			if (2 <= i)
			{
				z2 = (settings->thresholds[i-2] - node->pred_func)/sqrt(node->pred_var+settings->noisevar) ;
				phi2 = 0.5 + 0.5 * lerrf(z2) ;
			}
			if (trainlist->classes-1 >= i) 	
			{
				z1 = (settings->thresholds[i-1] - node->pred_func)/sqrt(node->pred_var+settings->noisevar) ;
				phi1 = 0.5 + 0.5 * lerrf(z1) ;
			}
			prob[i-1] = phi1 - phi2 ;  
			if (prob[i-1]>guess)
			{
				guess = prob[i-1] ;
				node->pred_label = i ;
			}
		}
		if (node->pred_label != node->target)
		{
			error += 1 ;
			abserr += abs(node->pred_label-node->target) ;
			if (prob[node->target-1]>0)
				logprob -= log(prob[node->target-1]) ;
			else
				logprob -= log(EPS) ;
		}
		node=node->next ;
	}
	free(prob) ;
	//return logprob ;
	return (double)error ;
}

int Dumping_Ordinal_Settings (Ordinal_Settings * settings)
{
	FILE * fid ;
	char buf[2048] ;
	unsigned int i ;

	if (NULL == settings)
		return 1 ;

	sprintf(buf,"%s.log",settings->pairs->filename) ;
	fid = fopen (buf,"w+t") ;
	if (NULL != fid)
	{
		// add testing information
		sprintf(buf, "final settings\n") ;
		fwrite (buf, sizeof(char), strchr(buf,'\0')-buf, fid) ;
		fprintf(fid,"kappa_o = %f\n",settings->kappa_o) ;
		if (0 == settings->ardon)
			fprintf(fid,"kappa = %f\n",settings->kappa_a) ;
		else 
		{
			for (i=0;i<settings->pairs->dimen;i++)
				fprintf(fid,"kappa%d = %f\n", i+1, settings->kappa[i]) ;				
		}
		fprintf(fid,"kappa_m = %f\n", settings->kappa_m) ;
		fprintf(fid,"noise variance = %f\n", settings->noisevar) ;
		for (i=0;i<settings->pairs->classes-1;i++)
			fprintf(fid,"threshold%d = %f\n", i+1, settings->thresholds[i]) ;
		fclose(fid) ;
	}
	return 0 ;
}

// end of gpor_predict.cpp

