// main function 

// Chu Wei (C) Copyright 2004 IPAM & GATSBY

#include <stdio.h>
#include <stdlib.h>
#include <math.h>
#include <limits.h>
#include <string.h>
#include <float.h>
#include "f2c.h"
#include "ordinal.h"
#define VERSION (1)
#define TRIALS (1)
#define TOPGENES (50)

int main( int argc, char * argv[] )
{
	Data_List rawdatalist ;
	Ordinal_Settings * settings ;
	double linear = 0 ;
	double kappa_o = 0 ;
	double kappa = 0 ;
	double kappa_m = 0 ;
	double regular = -1 ;
	unsigned int sz = 0 ;
	unsigned int index = 0 ;
	double parameter = 0 ;
	char buf[LENGTH]="" ;
	char filename[2048]="" ;
	char task[2048]="" ;
	//double points[TRIALS]={1};
	FILE * fid ;
	double sum = 0 ;
	int loop ;

	printf("\n Gaussian Processes for Gene Expression Analysis - version 1.%d\n - Chu Wei Copyright(C) 2004 at Gatsby Unit\n\n", VERSION) ;

	if ( 1 == argc )
	{
		// display help 
		printf(" Usage    : gpor [-K k] [-M m] [-O o][-R r] datafile\n") ;	
		printf(" - datafile:\t the data file of ordinal regression\n") ;
		printf(" - K k:\t set initial Kappa at k (default 1)\n") ;
		printf(" - M m:\t set initial kappa_M at m (default 0)\n") ;
		printf(" - O o:\t set initial kappa_O at o (default 1)\n") ;
		printf(" - L  :\t set Linear kernel (default Gaussian)\n") ;
		printf(" - R r:\t set Regularization factor at r (default 1)\n\n") ;
		return 0 ;
	}
	else
	{
		sprintf(task,"%s",argv[--argc]) ;
		if (argc>1)
			printf("Options:\n") ;
		do
		{
			strcpy(buf, argv[--argc]) ;
			sz = strlen(buf) ;
			if ( '-' == buf[0] )
			{			
				for (index = 1 ; index < sz ; index++)
				{
					switch (buf[index])
					{
					case '-' :
						break ;
					case 'K' :						
						if (parameter > 0)
						{ 
							if (index + 1 == sz)
							{
								printf("  - initialize Kappa at the value %f.\n", parameter) ;
								kappa = parameter ;
								parameter = 0;
							}
						}
						break ;
					case 'O' :						
						if (parameter > 0)
						{ 
							if (index + 1 == sz)
							{
								printf("  - initialize kappa_O at the value %f.\n", parameter) ;
								kappa_o = parameter ;
								parameter = 0;
							}
						}
						break ;
					case 'M' :
						if (parameter > 0)
						{ 
							if (index + 1 == sz)
							{
								printf("  - initialize kappa_M at the value %f.\n", parameter) ;
								kappa_m = parameter ;
								parameter = 0;
							}
						}
						break ;
					case 'L' :
						printf("  - use LINEAR kernel.\n") ;
						linear = 1 ;
						break ;
					case 'R' :						
						if (parameter >= 0)
						{ 
							if (index + 1 == sz)
							{
								printf("  - initialize Regularization factor at the value %f.\n", parameter) ;
								regular = parameter ;
								parameter = 0;
							}
						}
						break ;
					default :
						printf("  - %c is invalid.\n", buf[index]) ;
						break ;
					}
				}
			}
			else
				parameter = atof(buf) ;
		}
		while ( argc > 1 ) ;
		printf("\n") ;
	}

	// clean loo log
	sprintf(buf,"%s.loo",task) ;
	fid = fopen(buf,"w+t") ;
	fclose(fid) ;

	for (loop=0;loop<TOPGENES;loop++)
	{
		Create_Data_List (&rawdatalist) ;
		sprintf(filename,"%s.%d",task,loop+1) ;
		if ( 1 == ordinal_Loadfile (&rawdatalist, filename, 0) )
		{
			printf("fail to load data from %s.\n\n",filename) ;
			exit(1) ;
		}
		settings = Create_Ordinal_Settings(&rawdatalist) ;
		if (NULL == settings)
		{
			printf("fail to generate ordinal settings.\n") ;
			exit(1) ;
		}
		//settings->kappa_a = 1.0/(sqrt((double)rawdatalist.dimen)) ;
		Ordinal_EP_Training (settings) ;
		sum = Ordinal_EP_Loo (settings) ;
		Clear_Ordinal_Settings(settings) ;

		printf("leave one out error : %f\n" , sum) ;

		// save loo log
		sprintf(buf,"%s.loo",task) ;
		fid = fopen(buf,"a+t") ;
		fprintf(fid,"%f\n",sum) ;
		fclose(fid) ;
		
		Clear_Data_List (&rawdatalist) ;
	}

	return 0 ;
}


