#include <stdio.h>
#include <stdlib.h>
#include <math.h>
#include <string.h>
#include <time.h>
#include <assert.h>
#include "ordinal.h"
#include "neal_matrix.h"


double normcdf(double ca, double eb)
{ 
    int m,n,i,k;
        double eps ;
    double y[20],h,ep,p,x,s,q=0,a,b ;

        eps = EPS*EPS ;
        if (ca>eb)
        {       a=eb; b=ca;     }
        else
        {       a=ca; b=eb;     }

    h=b-a;
    y[0]=h*(normpdf(a)+normpdf(b))/2.0;
    m=1; n=1; ep=eps+1.0;
    while ((ep>=eps)&&(m<=19))
      { p=0.0;
        for (i=0;i<=n-1;i++)
          { x=a+(i+0.5)*h;
            p=p+normpdf(x);
          }
        p=(y[0]+h*p)/2.0;
        s=1.0;
        for (k=1;k<=m;k++)
          { s=4.0*s;
            q=(s*p-y[k-1])/(s-1.0);
            y[k-1]=p; p=q;
          }
        ep=fabs(q-y[m-1]);
        m=m+1; y[m-1]=q; n=n+n; h=h/2.0;
      }
        if (ca>eb)
    return(-q) ;
        else
    return(q) ;
}



double Retrieve_Posterior_Covariance (unsigned int i, unsigned int j, Ordinal_Settings * settings)
{
	if (NULL == settings)
		return 0 ;
	if (i>=settings->pairs->count||j>=settings->pairs->count)
	{
		printf("index is out of range.\n") ;
		return 0 ;
	}
	if (i >= j)
		return (settings->alpha+i)->postcov[j] ;
	else 
		return (settings->alpha+j)->postcov[i] ; 
}

double Ordinal_EP_Loo(Ordinal_Settings * settings)
{
	double loo = 0, Aii ;
	unsigned int index ;
	Alphas * alpha ;
	Data_Node * node ;
	
	assert(settings) ;
	for (index=0;index<settings->pairs->count;index++)
	{
		// leave index out
		Aii = Retrieve_Posterior_Covariance (index, index, settings) ;
		alpha = settings->alpha+index ;
		node = alpha->pair ;		
		alpha->loovar = Aii/(1-Aii*node->epinvvar) ;
		alpha->loomean = node->postmean + alpha->loovar*node->epinvvar*(node->postmean-node->epmean) ;
		//if ((alpha->loomean-settings->thresholds[0])*(alpha->pair->target-1.5)<0)
		//	loo+=1 ;
		Aii = Phi(settings->thresholds[0] - alpha->loomean)/sqrt(alpha->loovar+settings->noisevar) ;
		if (Aii>1.0-EPS)
			Aii=1.0-EPS ;
		if (Aii<EPS)
			Aii=EPS ;
		if (alpha->pair->target<1.5)
			loo -= log(Aii);
		else
			loo -= log(1-Aii);
	}
	return loo;
}

double Ordinal_EP_Swipe(Ordinal_Settings * settings)
{
	double error = 0 ;
	unsigned int index ;
	double Aii, Zi ;
	double z1, z2, phi1, phi2, n1, n2, diff, rho, eta, dphi ;
	Alphas * alpha ;
	Data_Node * node ;
	double * ai ;
	unsigned int i , j, s=0 ;
	int * flag ;

	srand( (unsigned)time( NULL ) );
	//srand(2004) ;
	flag = (int *) calloc( settings->pairs->count,sizeof(int) ) ;
	//for (index=0;index<settings->pairs->count;index++)
	while (s<settings->pairs->count)
	{
		do 
		{
			index=(unsigned int)floor((double)rand()/(RAND_MAX+1.0)*settings->pairs->count) ;
		}
		while (flag[index]>0) ;
		flag[index] = 1 ;
		s += 1 ;
		// leave index out
		Aii = Retrieve_Posterior_Covariance (index, index, settings) ;
		alpha = settings->alpha+index ;
		node = alpha->pair ;
		
		alpha->loovar = Aii/(1-Aii*node->epinvvar) ;
		alpha->loomean = node->postmean + alpha->loovar*node->epinvvar*(node->postmean-node->epmean) ;

		// compute Z
		phi2 = 0 ;
		phi1 = 1 ;
		n1 = 0 ;
		n2 = 0 ;
		z1=0 ; 
		z2=0 ;

		if (1 == node->target)
		{
			z1 = (settings->thresholds[node->target-1] - alpha->loomean)/sqrt(alpha->loovar+settings->noisevar) ;
			n1 = normpdf(z1) ;
			phi1 = Phi(z1) ;
			dphi = phi1-phi2 ;

		}
		else if (settings->pairs->classes == node->target)
		{
			z2 = (settings->thresholds[node->target-2] - alpha->loomean)/sqrt(alpha->loovar+settings->noisevar) ;
			n2 = normpdf(z2) ;
			phi2 = Phi(z2) ;
			dphi = phi1-phi2 ;
		}
		else
		{
			z1 = (settings->thresholds[node->target-1] - alpha->loomean)/sqrt(alpha->loovar+settings->noisevar) ;
			z2 = (settings->thresholds[node->target-2] - alpha->loomean)/sqrt(alpha->loovar+settings->noisevar) ;
			n1 = normpdf(z1) ;
			n2 = normpdf(z2) ;
			dphi = Phi(z1) - Phi(z2) ;
		}
		
		if(dphi<0) 
			dphi = normcdf(z2,z1) ;

		Zi = dphi ;
		if (Zi<EPS)
		{
			if ( fabs(exp(-0.5*z1*z1+0.5*z2*z2)-1.0)>EPS*EPS )
			{
			alpha->alpha = (z1*exp(-0.5*z1*z1+0.5*z2*z2)-z2)/(exp(-0.5*z1*z1+0.5*z2*z2)-1.0)/sqrt(alpha->loovar+settings->noisevar) ;
			alpha->beta = (-1.0+(z1*z1*exp(-0.5*z1*z1+0.5*z2*z2)-z2*z2)/(exp(-0.5*z1*z1+0.5*z2*z2)-1.0))/(alpha->loovar+settings->noisevar)/2.0 ;
			alpha->nu = alpha->alpha*alpha->alpha-2.0*alpha->beta ;
			}
			else
			{
				alpha->alpha = 0 ;
				alpha->beta = -(1.0-EPS)/(alpha->loovar+settings->noisevar)/2.0 ;
				alpha->nu = (1.0-EPS)/(alpha->loovar+settings->noisevar) ;
				printf("get in approximate dphi, a=%f b=%f nu=%f\n",alpha->alpha, alpha->beta, alpha->nu*(alpha->loovar+settings->noisevar)) ;
			}
			if (alpha->nu>=1.0/(alpha->loovar+settings->noisevar))
				alpha->nu = (1.0-EPS)/(alpha->loovar+settings->noisevar) ;
			if (alpha->nu<=0)
				alpha->nu = EPS*(alpha->loovar+settings->noisevar) ;
#ifdef _ORDINAL_DEBUG
			printf("get in approximate dphi, a=%f b=%f nu=%f\n",alpha->alpha, alpha->beta, alpha->nu*(alpha->loovar+settings->noisevar)) ;
#endif
		}
		else
		{
			alpha->alpha = -(n1-n2)/Zi/sqrt(alpha->loovar+settings->noisevar) ;
			alpha->beta = -(z1*n1-z2*n2)/Zi/(alpha->loovar+settings->noisevar)/2.0 ;
			alpha->nu = alpha->alpha*alpha->alpha-2.0*alpha->beta ;
		}
#ifdef _ORDINAL_DEBUG
		//printf("%d :  Z=%f \n",index,Zi) ;		
		//printf("%d :  a=%f \n",index,alpha->alpha) ;	
		//printf("%d :  b=%f \n",index,alpha->beta) ;	
		assert(alpha->nu>0) ;
		assert(alpha->nu<1.0/(alpha->loovar+settings->noisevar)) ;
#endif
		alpha->hnew = alpha->loomean + alpha->loovar * alpha->alpha ;
		alpha->pnew = alpha->nu/(1.0-alpha->loovar*alpha->nu) ;
		alpha->mnew = alpha->loomean + alpha->alpha/alpha->nu ;
		alpha->snew = Zi*sqrt(alpha->loovar*alpha->pnew+1.0)*exp(0.5*alpha->alpha*alpha->alpha/alpha->nu) ;
#ifdef _ORDINAL_DEBUG
		assert(alpha->pnew>0) ;
#endif

		// update posterior mean and cov
		diff = alpha->pnew - node->epinvvar ;
		if (fabs(diff)>(EPS*EPS)&&alpha->pnew>0&&Zi>0)
		{
			rho = diff/(1+diff*Aii) ;
			eta = (alpha->alpha+node->epinvvar*(node->postmean-node->epmean))/(1.0-Aii*node->epinvvar) ;
			// prepare the ai vector ;
			ai = (double *) malloc( settings->pairs->count*sizeof(double) ) ;
			if (NULL==ai)
			{
				printf("fail to malloc ai\n") ;
				exit(1) ;
			}
			for (i=0;i<settings->pairs->count;i++)
				ai[i] = Retrieve_Posterior_Covariance (i, index, settings) ;
			// update posterior cov and mean
			for (i=0;i<settings->pairs->count;i++)
			{
				for (j=0;j<=i;j++)
				{
					(settings->alpha+i)->postcov[j]-=rho*ai[i]*ai[j] ;
				}
				(settings->alpha+i)->pair->postmean+=eta*ai[i] ;
			}
			free(ai) ;
			// update epmean epvar epamp
			error += (alpha->pnew-node->epinvvar)*(alpha->pnew-node->epinvvar) ;
			error += (alpha->mnew-node->epmean)*(alpha->mnew-node->epmean) ;
			error += (alpha->snew-node->epamp)*(alpha->snew-node->epamp) ;
			node->epinvvar = alpha->pnew ;
			node->epmean = alpha->mnew ;
			node->epamp = alpha->snew ;
#ifdef _ORDINAL_DEBUG
		//printf("%d :  epm=%f \n",index,node->epmean) ;		
		//printf("%d :  epp=%f \n",index,node->epinvvar) ;	
		//printf("%d :  eps=%f \n",index,node->epamp) ;	
#endif
		}
		else
		{	
			//error += EPS ;
			if (alpha->pnew<0)
				printf("skip %d sample --- z1 %f z2 %f nu %f \t %f * %f.\n", index+1,
                                	z1, z2, alpha->nu, alpha->pnew, node->epinvvar) ;

#ifdef _ORDINAL_DEBUG
			printf("skip %d sample --- z1 %f z2 %f nu %f \t %f * %f.\n", index+1, 
				z1, z2, alpha->nu, alpha->pnew, node->epinvvar) ;		
			
#endif		
		}
	}
	free(flag) ;
	return error ;
}

int Compute_EP_Weights (Ordinal_Settings * settings) 
{
	double * t1 ;
	double * t2 ;
	double temp ;
	unsigned int i , j ;

	if (NULL == settings)
		return 1 ;

	for (i=0;i<settings->pairs->count;i++)
	{
		if (0==(settings->alpha+i)->pair->epinvvar)
		{
			printf("%u sample has not been updated.\n",i+1) ;
			return 1 ;
		}
		else
			(settings->alpha+i)->pair->weight = (settings->alpha+i)->alpha ;
	}
	
	if (NULL != settings->invcov)
		free(settings->invcov) ;

	settings->invcov = (double*)malloc(settings->pairs->count*settings->pairs->count*sizeof(double)) ;
	t1 = (double*)malloc(settings->pairs->count*sizeof(double)) ;
	t2 = (double*)malloc(settings->pairs->count*sizeof(double)) ;
	for (i=0;i<settings->pairs->count;i++)
	{	
		for (j=0;j<i;j++)
		{
			settings->invcov[i*settings->pairs->count+j] = (settings->alpha+i)->kernels[j] ;
			settings->invcov[j*settings->pairs->count+i] = settings->invcov[i*settings->pairs->count+j] ;
		}
		settings->invcov[i*settings->pairs->count+i] = (settings->alpha+i)->kernels[i] ;
		settings->invcov[i*settings->pairs->count+i] += 1.0/(settings->alpha+i)->pair->epinvvar ;
	}
	if (0 == cholesky(settings->invcov,settings->pairs->count,&temp))
		printf("Fatal Error: Sigma+1/Pi is not positive definite.\n") ;
	if (0 == inverse_from_cholesky(settings->invcov, t1, t2, settings->pairs->count) )		
		printf("Fatal Error: Sigma+1/Pi is not positive definite.\n") ;
	
	for (i=0;i<settings->pairs->count;i++)
	{
		(settings->alpha+i)->pair->weight = 0 ;
		for (j=0;j<settings->pairs->count;j++)
			(settings->alpha+i)->pair->weight += settings->invcov[i*settings->pairs->count+j]*(settings->alpha+j)->pair->epmean ;
		// check the equilibrium
		if (fabs((settings->alpha+i)->alpha-(settings->alpha+i)->pair->weight)>sqrt(EPS))
		{
			printf("\nFatal error : %u sample is not at equilibrium %f ---- %f.\n",i+1,(settings->alpha+i)->alpha,(settings->alpha+i)->pair->weight) ;
			exit(1) ;
		}
	}
	free(t1);
	free(t2);
	return 0 ;	
}

int Ordinal_EP_Training (Ordinal_Settings * settings) 
{
	double error ;
	int index = 0 ;
#ifdef _ORDINAL_DEBUG
	unsigned int i=0 ;
#endif

	if (NULL == settings)
		return 0 ;

#ifdef _ORDINAL_DEBUG
	printf("Noise Variance: %.3f\n",settings->noisevar) ;
	printf("Kappa_O: %.3f\n",settings->kappa_o) ;
	printf("Thresholds:\n") ;
	for (i=1;i<settings->pairs->classes;i++)
		printf("%u ---  %.3f\n",i,settings->thresholds[i-1]) ;
	if (0==settings->ardon)		
		printf("Kernel: %.3f\n",settings->kappa_a) ;
	else if (settings->pairs->dimen<20)
	{
		printf("Kernel:\n") ;
		for (i=0;i<settings->pairs->dimen;i++)
			printf("%u ---  %.6f\n", i, settings->kappa[i]) ;
	}
#endif
	// may refresh ALPHAS here 
	if (NULL != settings->alpha)
		Clear_Alphas(settings->alpha, settings->pairs) ;
	settings->alpha = Create_Alphas (settings) ;
	if (NULL == settings->alpha)
	{
		printf("fail to create ALPHAs.\n") ;
		exit(1) ;
	}
	do
	{
		error = Ordinal_EP_Swipe(settings) ; // clear sign
#ifdef _ORDINAL_DEBUG
		printf("Loop %d --- %f\n", index, error) ;
#endif
		index += 1 ;
	}
	while (error/(double)settings->pairs->count>EPS) ;

	// check the equilibrium
	if (1 == Compute_EP_Weights (settings))
		return 1 ;

#ifdef _ORDINAL_DEBUG
	for (index=0;index<(int)settings->pairs->count;index++)

		printf("%u --- post m %.3f - m %.3f -p %.3f -s %.3f -a %.3f -ai %.3f -vi %.3f\n", 
		(settings->alpha+index)->pair->target, 
		(settings->alpha+index)->pair->postmean,
		(settings->alpha+index)->pair->epmean,
		(settings->alpha+index)->pair->epinvvar,
		(settings->alpha+index)->pair->epamp, 
		(settings->alpha+index)->pair->weight,
		(settings->alpha+index)->alpha,
		(settings->alpha+index)->nu) ;
#endif
	return 0 ;
}

