/*******************************************************************************\

	loadfile.c in Sequential Minimal Optimization ver2.0 
	
	loads data file from disk in data list.

	Chu Wei Copyright(C) National Univeristy of Singapore
	Create on Jan. 16 2000 at Control Lab of Mechanical Engineering 
	Update on Aug. 23 2001 

\*******************************************************************************/

//#include <conio.h>
#include <stdio.h>
#include <stdlib.h>
#include <string.h>
#include <time.h>
#include <math.h>
#include <limits.h>
#include "smo.h"

/*******************************************************************************\

	BOOL smo_Loadfile ( Pairs * pairs, char * filename, unsigned int inputdim ) 
	
	load data file settings->inputfile, and create the data list Pairs 
	input:  the pointers to pairs and filename
	output: TRUE or FALSE

\*******************************************************************************/

BOOL smo_Loadfile ( Data_List * pairs, char * inputfilename, int inputdim ) 
{ 

	FILE * smo_stream ;
	FILE * smo_target = NULL ;
	char * pstr = NULL ;
	char buf[LENGTH]="" ;
	char * temp ;
	int dim = -2 ;
	unsigned long index = 1 ;
	unsigned int result ;
	int var = 0, chg = 0 ;
	double * point = NULL ;
	double y ;
	int i = 0, j = 0 ;
	double mean = 0 ;
	double ymax = LONG_MIN ;
	double ymin = LONG_MAX ;
	double * xmean = NULL;
	Data_Node * node = NULL ;
	double y1 = 0, y2 = 0, y3 = 0 ;

	if ( NULL == pairs || NULL == inputfilename )
		return FALSE ;
	
	Clear_Data_List( pairs ) ;
	//Create_Data_List(pairs) ;
        if( (smo_stream = fopen( inputfilename, "r+t" )) == NULL )
	{
		//printf( "can not open the file %s.\n", inputfilename );
		return FALSE ;
	}

	// check the input dimension here
	if ( NULL == fgets( buf, LENGTH, smo_stream ))
	{
		printf( "fgets error in reading the first line.\n" );
		fclose( smo_stream );
		return FALSE ;
	}
	
	var = strlen( buf ) ;
	
	if (var >= LENGTH-1) 
	{
		printf( "the line is too long in the file %s.\n", inputfilename );
		fclose( smo_stream );
		return FALSE ;
	}
	
	if (0 < var)
	{		
		do 
		{
			dim = dim + 1 ;
			strtod( buf, &temp ) ;
			strcpy( buf, temp ) ;
			chg = var - strlen(buf) ;
			var = var - chg ;
		}
		while ( 0 != chg ) ;
	}
	else
	{ 
		fclose( smo_stream );
		printf("the first line in the file is empty.\n") ;
		return FALSE ;
	}

	if ( 0 > dim || (0 == dim && 0 == inputdim) ) 
	{
		fclose( smo_stream );

#ifdef SMO_DEBUG
		printf( "input dimension is less than one.\n") ;
#endif
		return FALSE ;
	}
	else
	{
		if (0 == inputdim)
			printf("\nDimension %d\n",dim);
	}	
	if (inputdim > 0)
	{
		if (inputdim == dim + 1 ) // test file without target
		{
			// try to open "*target*.*" as target
			// create target file name
			pstr = strstr( inputfilename , "test" ) ;
			if (NULL != pstr)
			{
				result = abs( inputfilename - pstr ) ;
				strncpy (buf, inputfilename, result ) ;
				buf[result] = '\0' ;
				strcat(buf, "targets") ;
				strcat (buf, pstr+4) ;
				smo_target = fopen( buf, "r+t" ) ;
			}
			dim = inputdim ;
			pairs->dimen = dim ;
		}
		else if ( inputdim != dim )
		{
			printf("Dimensionality in testdata is inconsistent with traindata.\n") ;
			return FALSE ;
		}
		else
			pairs->dimen = dim ;
	}
	else
		pairs->dimen = dim ;
	
	//initialize the x_mean and x_devi in Data_List pairs

	if ( NULL == (pairs->x_mean = (double *)(malloc(dim*sizeof(double))) ) 
		|| NULL == (pairs->x_devi = (double *)(malloc(dim*sizeof(double))) ) 
		|| NULL == (xmean = (double *)(malloc(dim*sizeof(double))) ) )
	{		
		if (NULL != pairs->x_mean) 
			free(pairs->x_mean) ;
		if (NULL != pairs->x_devi) 
			free(pairs->x_devi) ;
		if (NULL != xmean)
			free(xmean) ;
		if (NULL != smo_target)
			fclose( smo_target ) ;
		if (NULL != smo_stream)
			fclose( smo_stream );
		return FALSE ;
	}
	for ( j = 0; j < dim; j ++ )
		pairs->x_mean[j] = 0 ;
	for ( j = 0; j < dim; j ++ )
		pairs->x_devi[j] = 0 ;
	for ( j = 0; j < dim; j ++ )
		xmean[j] = 0 ;

	// begin to initialize data_list for digital input only
	printf("\nLoading %s ...  \n", inputfilename) ;
	pairs->datatype = REGRESSION ; 

	rewind( smo_stream ) ;
        fgets( buf, LENGTH, smo_stream ) ;
	do
	{
		//printf("%d\n", index) ;
		point = (double *) malloc( dim * sizeof(double) ) ; // Pairs to free them
		if ( NULL == point )
		{
			printf("not enough memory.\n") ;
			if (NULL != smo_target)
				fclose( smo_target ) ;
			if (NULL != smo_stream)
				fclose( smo_stream );
			if (NULL != pairs->x_mean) 
				free(pairs->x_mean) ;
			if (NULL != pairs->x_devi) 
				free(pairs->x_devi) ;
			if (NULL != xmean)
				free(xmean) ;
			Clear_Data_List( pairs ) ;
			return FALSE ;
		}
		var = strlen( buf ) ;	
		i = 0 ;
		chg = dim ;

		while ( chg>0 && i<dim)
		{
			point[i] = strtod( buf, &temp ) ;
			i++ ;
			strcpy( buf, temp ) ;
			chg = var - strlen(buf) ;
			var = var - chg ;
		}
		if (i==dim && chg>0 && var>0)
			y = strtod( buf, &temp ) ;
		else
		{
			free(point) ;
			y = 0 ;
			printf("Warning: the input file contains a blank or defective line.\n") ;
		}
		// load y as target from other file when dim+1
		if (NULL != smo_target)
		{
			if ( NULL != fgets( buf, LENGTH, smo_target ) )
			{
				var = strlen( buf ) ;
				y = strtod( buf, &temp ) ;
				strcpy( buf, temp ) ;
				chg = var - strlen(buf) ;
				if (chg==0)
					printf("Warning: the target file contains a blank line.\n") ;
			}
			else
				printf("Warning: the target file is shorter than the input file.\n") ;
		}

		/*	for ( i = 0; i < dim; i ++ )
			{
				point[i] = strtod( buf, &temp ) ;
				strcpy( buf, temp ) ;
			}
			y = strtod( buf, &temp ) ;

			// load y as target from other file when dim+1
			if (NULL != smo_target)
			{
				fgets( buf, LENGTH, smo_target ) ;
				y = strtod( buf, &temp ) ;
			}*/

		if (chg>0) 
		{	
						
			if ( TRUE == Add_Data_List( pairs, Create_Data_Node(index, point, y) ) )
			{
				// update statistics
				//printf("point number : %ld\n",pairs->count);
                                pairs->mean = (mean * (((double)(pairs->count)) - 1) + y )/ ((double)(pairs->count))  ;
				pairs->deviation = pairs->deviation + (y-mean)*(y-mean) * ((double)(pairs->count)-1)/((double)(pairs->count));			
				mean = pairs->mean ;	
				for ( j=0; j<dim; j++ )
				{
					pairs->x_mean[j] = (xmean[j] * (((double)(pairs->count)) - 1) + point[j] )/ ((double)(pairs->count))  ;
					pairs->x_devi[j] = pairs->x_devi[j] + (point[j]-xmean[j])*(point[j]-xmean[j]) * ((double)(pairs->count)-1)/((double)(pairs->count));			
					xmean[j] = pairs->x_mean[j] ;
				}
				if (y>ymax)
				{ ymax = y ; pairs->i_ymax = index ;}
				if (y<ymin)
				{ ymin = y ; pairs->i_ymin = index ;}
				
				// check data type 
				if ( CLASSIFICATION == pairs->datatype || UNKNOWN == pairs->datatype )
				{
					if ( 1 == index )
					{
						y1 = y ;
						pairs->datatype = UNKNOWN ;
					}
					else if ( 0 == y3 && y != y1 )
					{
						y3 = 1 ;
						y2 = y ;
						pairs->datatype = CLASSIFICATION ;
					}
					else if ( y != y1 && y != y2 )
						pairs->datatype = REGRESSION ;
				}
				index ++ ;
			}
			else
			{
				printf("%d\n", (int)index) ;
				printf("duplicate data \n") ;
			}
		}
	}
	while( !feof( smo_stream ) && NULL != fgets( buf, LENGTH, smo_stream ) ) ;


	if (pairs->count < MINNUM || (pairs->datatype == UNKNOWN && inputdim == 0 ) ) 
	{
		if(pairs->count<MINNUM)
                        printf("too few input pairs: %ld\n",pairs->count) ;
		Clear_Data_List( pairs ) ;
		if (NULL != pairs->x_mean) 
			free(pairs->x_mean) ;
		if (NULL != pairs->x_devi) 
			free(pairs->x_devi) ;
		if (NULL != xmean)
			free(xmean) ;
		if (NULL != smo_target)
			fclose( smo_target ) ;
		if (NULL != smo_stream)
			fclose( smo_stream );
		return FALSE ;
	}

	pairs->deviation = sqrt( pairs->deviation / ((double)(pairs->count - 1.0)) ) ;
	for ( j=0; j<dim; j++ )
	{
		pairs->x_devi[j] = sqrt( pairs->x_devi[j] / ((double)(pairs->count - 1.0)) ) ;	

	}
	
	// set target value as +1 or -1, if data type is CLASSIFICATION
	if ( CLASSIFICATION == pairs->datatype && 0 == inputdim )
	{
			node = pairs->front ;
			while ( node != NULL )
			{
				if (y1 > y2)
					node->target = (y1 == node->target) ? +1.0 : -1.0 ;
				else if (y1 < y2)
					node->target = (y1 == node->target) ? -1.0 : +1.0 ;
				else
					pairs->datatype = UNKNOWN ;
				node = node->next ; 
			}
			pairs->deviation = 1.0 ;
			pairs->mean = 0 ;
			pairs->normalized_output = FALSE ;
	}

	if (inputdim>0) // do not normailize data for TESTING
	{
		pairs->normalized_output = FALSE ;
		pairs->normalized_input = FALSE ; 
	}

	// normalize the target if needed 
	node = pairs->front ;
	while ( node != NULL )
	{
		if ( TRUE == pairs->normalized_output )
		{
			if (pairs->deviation>0)
				node->target = (node->target - pairs->mean)/(pairs->deviation) ;
			else
				node->target = 0 ;
		}
		if ( TRUE == pairs->normalized_input )
		{
			for ( j=0; j<dim; j++ )
			{				
				if (pairs->x_devi[j]>0)
					node->point[j] = (node->point[j]-pairs->x_mean[j])/(pairs->x_devi[j]) ;
				else
					node->point[j] = 0 ;
			}
		}
		node = node->next ; 
	}
	printf("Total %d training samples with %d dimensions for ", (int)pairs->count, (int)pairs->dimen) ;	

	if	(inputdim > 0)
		printf("TESTING.\r\n") ;
	else 
	{		
		if( CLASSIFICATION == pairs->datatype )
			printf("CLASSIFICATION.\r\n") ;
		else
			printf("REGRESSION.\r\n") ;
	}
	if (TRUE == pairs->normalized_input)
		printf("Inputs are normalized.\r\n") ;
	
	if (TRUE == pairs->normalized_output && pairs->deviation > 0)
		printf("Outputs are normalized.\r\n") ;
	if ( inputdim > 0 && pairs->deviation <= 0 )
		printf("Targets are not at hand.\r\n") ;

	if (NULL != smo_target)
		fclose( smo_target ) ;
	//if (NULL != smo_stream)
		fclose( smo_stream );
	if ( NULL != xmean )
		free( xmean ) ;
	return TRUE ;
}


// end of loadfile.c
