/*******************************************************************************\

	main.c to generate Feature Ranking Profiles ver2.0
		
	entry function.
		
	Chu Wei Copyright(C) Gatsby Unit

\*******************************************************************************/

#include <stdio.h>
#include <stdlib.h>
#include <malloc.h>
#include <string.h>
#include <limits.h>
#include <math.h>
#include <float.h>
#include <time.h>
#include "smo.h"
#define VERSION (0)
#define TOPGENES (50)
#define KFOLDS (10)

double mcnemar(long int samples,int * validoutput,int * baseline, int * target) 
{
	long int n00=0, n11=0, n10=0, n01=0 ;
	long int i ;
	double x2 ;
	double hold = 3.841459 ;
	if (NULL==validoutput||NULL==baseline||NULL==target)
		return 0 ;
	for (i=0;i<samples;i++)
	{
		if (validoutput[i]==target[i]&&baseline[i]==target[i])
			n11+=1 ;
		else if (validoutput[i]!=target[i]&&baseline[i]!=target[i])
			n00+=1 ;
		else if (validoutput[i]==target[i]&&baseline[i]!=target[i])
			n10+=1 ;
		else if (validoutput[i]!=target[i]&&baseline[i]==target[i])
			n01+=1 ;
	}
	if (n01+n10<=0)
		return -1.0 ;
	x2 = (double)((fabs(n01-n10)-1)*(fabs(n01-n10)-1))/(double)(n01+n10) ;
	printf("X2 = %f\n",x2) ;
	//return x2 ;
	if (x2>hold)
		return x2 ;
	else
	{
		if (n01<n10)
		{
			//update baseline
			for (i=0;i<samples;i++)
				baseline[i] = validoutput[i] ;	
			return -x2 ;		
		}
		else
			return x2 ;
	}
}

int main( int argc, char * argv[])
{
	def_Settings * defsetting = NULL ;
	Data_Node * node ;
	//char task[1024] ;
	char buf[LENGTH] ;
	FILE * log ; 
	double temp ;
	unsigned int index, subset=0, i, folds=0 ;
	unsigned long * rank ;
	float * hits = NULL ;
	unsigned int * uhits = NULL ;
	double * error ;
	float ardsum = 0 ;
	int * feature ;
	float * ard ;
	
	printf("\nGenerating Selected Features v1.%d \n --- Chu Wei Copyright(C) 2004\n", VERSION) ;
	if ( 1 == argc || NULL == (defsetting = Create_def_Settings(argv[--argc])) )
	{
		// display help	
		printf("\nUsage: gtops file \n\n") ;
		printf("  file   specifies the file with sorted features.\n") ;
		printf("\n") ;
		if (NULL !=defsetting)
			Clear_def_Settings( defsetting ) ;
		return 0 ;
	}

if ( TRUE == Update_def_Settings(defsetting) ) 
{
	hits = (float *) calloc(defsetting->pairs.dimen+1,sizeof(float)) ;
	uhits = (unsigned int *) calloc(defsetting->pairs.dimen+1,sizeof(unsigned int)) ;
	// load the leave-one-out error.
	for (folds=1;folds<=KFOLDS;folds++) 
	{
		error = (double *) calloc(TOPGENES,sizeof(double)) ;
		sprintf(buf,"%s_fold.%u.loo",argv[argc],folds) ;
		log = fopen(buf,"r+t") ;
		if (NULL != log)
		{	
			i=0;
			printf("load %s ...\n",buf) ; 
			while (!feof(log)&&NULL!=fgets(buf,LENGTH-1,log))
			{
				i+=1 ;
			}
			printf("Sample Size: %d\n",i) ;
			if (i!=TOPGENES)
				printf("Warning : the number of samples is wrong.\n") ;
			rewind(log) ;
			i=0;
			while (!feof(log)&&NULL!=fgets(buf,LENGTH-1,log))
			{			
				error[i] = atof(buf) ;
				i+=1 ;
			}
			fclose(log) ;
		}
		else
		{
			printf("fail to open %s.\n",buf) ;
			//exit(1) ;
		}

		// load the top features.
		sprintf(buf,"%s_fold.%u.feat",argv[argc],folds) ;
 		feature = (int *) calloc(TOPGENES,sizeof(int)) ;
                ard = (float *) calloc(TOPGENES,sizeof(float)) ;
		ardsum = 0 ;
		log = fopen(buf,"r+t") ;
		if (NULL != log)
		{	
			i=0;
			printf("load %s ...\n",buf) ;
			while (!feof(log)&&NULL!=fgets(buf,LENGTH-1,log))
			{
				i+=1 ;
			}
			printf("Sample Size: %d\n",i) ;
			if (i!=TOPGENES)
				printf("Warning : the number of samples is wrong.\n") ;
			rewind(log) ;
			i=0;
			while (!feof(log)&&NULL!=fgets(buf,LENGTH-1,log))
			{			
				sscanf(buf,"%d %f",&(feature[i]),&(ard[i]));
				ardsum += ard[i] ;
				i+=1 ;
			}
			fclose(log) ;
		}
		else
		{
			printf("fail to open %s.\n",buf) ;
			//exit(1) ;
		}

		if (ardsum>0)
		{
		temp = error[0] ;
		subset = 1 ;
		for (i=1;i<TOPGENES;i++)
		{	
			if (temp>=error[i])
			{
				temp = error[i] ;
				subset = (int)i+1 ;
			}
		}

		printf("%d genes are selected.\n", subset) ;
		for (i=0;i<subset;i++)
		{
			hits[feature[i]] += (float) (1.0+ard[i]/ardsum) ;
			//hits[feature[i]] += (float) (ard[i]/ardsum) ;
			uhits[feature[i]] += 1 ;
			//printf("gene %d\n",feature[i]) ;
		}	
		}
		free(ard) ;
		free(feature) ;
		free(error) ;
	}

	rank = (unsigned long *) malloc((defsetting->pairs.dimen+1)*sizeof(unsigned long)) ;
	indexx(defsetting->pairs.dimen,hits,rank) ;
	
	//for (index=0;index<=defsetting->pairs.dimen;index++)
	//	printf("%f -- %lu\n", hits[index],rank[index]) ;
	folds = 0 ;
	while(folds<TOPGENES)
	{
		// generate sub files for LOO
		printf("%u : %f-%lu\n",folds+1,hits[rank[defsetting->pairs.dimen-folds]],rank[defsetting->pairs.dimen-folds]) ;
		sprintf(buf,"%s_sort.%u",defsetting->inputfile,folds+1) ;
		log = fopen(buf,"w+t") ;
		//printf("generate %s ... \n",buf) ;
		if (NULL != log)
		{
			node = defsetting->pairs.front ;
			while (NULL != node)
			{
				// each point
				for (index = 0 ; index <= folds ; index++)
				{
					fprintf(log,"%f ",node->point[rank[defsetting->pairs.dimen-index]-1]) ; 
				}
				fprintf(log,"%.0f\n", node->target) ; 
				node = node->next ;
			}
			fclose(log) ;		
		}
		folds += 1 ;
	}

	// generate top features
	sprintf(buf,"%s.topfeat",defsetting->inputfile) ;
	log = fopen(buf,"w+t") ;
	if (NULL != log)
	{
		for (index = 0 ; index < TOPGENES ; index++)
	    	{
			fprintf(log,"%lu %u\n",rank[defsetting->pairs.dimen-index],uhits[rank[defsetting->pairs.dimen-index]]) ; 
			//printf("%u, gene %lu\n",index+1,rank[defsetting->pairs.dimen-index]) ;
		}
		fclose(log) ;		
	}
	free(rank) ;
	free(hits) ;
	free(uhits) ;
}
	Clear_def_Settings( defsetting ) ;
	return 0;
}
//end of main.c 
