#include <stdio.h>
#include <stdlib.h>
#include <math.h>
#include "ordinal.h"
#include "neal_matrix.h"

int Ordinal_EPEM_Evaluate_FuncGrad ( void * pointer, bfgs_FuncGrad * bfgs ) 
{
	Ordinal_Settings * settings ;
	Alphas * alpha ;
	Alphas * alphab ;
	double * t1 ;
	double * t2 ;
	double * t3 ;
	double * t4 ; // for noise
	double * t5 ;
	double * dsigma ;
	double temp ;
	double cbe ;
	unsigned int i ;
	unsigned int j ;
	unsigned int ii ; // settings->kappa


	if (NULL == pointer || NULL == bfgs)
		return 1 ;
	settings = pointer ;

	bfgs_ordinal_Settings( settings,settings->bfgs ) ;

	// call EP routine to find posterior distribution 
	if (1 == Ordinal_EP_Training (settings) )
	{
		printf("\nEP failed to get equilibrium.\n\n") ;
		//return 1 ;
	}

	// evaluate functional and gradient
	if (NULL != settings->invcov)
		free(settings->invcov) ;

	settings->invcov = (double*)malloc(settings->pairs->count*settings->pairs->count*sizeof(double)) ;
	dsigma = (double*)malloc(settings->pairs->count*settings->pairs->count*sizeof(double)) ;
	t1 = (double*)malloc(settings->pairs->count*sizeof(double)) ;
	t2 = (double*)malloc(settings->pairs->count*sizeof(double)) ;
	t3 = (double*)malloc(settings->pairs->count*sizeof(double)) ;
	t4 = (double*)malloc(settings->pairs->count*sizeof(double)) ;
	t5 = (double*)malloc(settings->pairs->count*sizeof(double)) ;

	bfgs->fx = 0 ;
	for (i=0;i<settings->pairs->count;i++)
	{	
		for (j=0;j<i;j++)
		{
			settings->invcov[i*settings->pairs->count+j] = (settings->alpha+i)->kernels[j] ;
			settings->invcov[j*settings->pairs->count+i] = settings->invcov[i*settings->pairs->count+j] ;
		}
		settings->invcov[i*settings->pairs->count+i] = (settings->alpha+i)->kernels[i] ;
		settings->invcov[i*settings->pairs->count+i] += 1.0/(settings->alpha+i)->pair->epinvvar ;
		bfgs->fx -= 0.5 * (settings->alpha+i)->pair->postmean * (settings->alpha+i)->pair->weight ;
	}
	if (0 == cholesky(settings->invcov,settings->pairs->count,&temp))
		printf("Fatal Error: Sigma+1/Pi is not positive definite.\n") ;

	for (i=0;i<settings->pairs->count;i++)
	{
		bfgs->fx -= log(settings->invcov[i+i*settings->pairs->count]) ;
		if ((settings->alpha+i)->pair->epinvvar>0)
			bfgs->fx -= 0.5*log((settings->alpha+i)->pair->epinvvar);
		else
			bfgs->fx -= 0.5*log(EPS*EPS);
	}

	if (0 == inverse_from_cholesky(settings->invcov, t1, t2, settings->pairs->count) )		
		printf("Fatal Error: Sigma+1/Pi is not positive definite.\n") ;

	for (i=0;i<settings->pairs->count;i++)
	{
		if ((settings->alpha+i)->pair->epinvvar>0)
			bfgs->fx -= 0.5 * settings->invcov[i+i*settings->pairs->count] / (settings->alpha+i)->pair->epinvvar ;	
		else
			bfgs->fx -= 0.5 * settings->invcov[i+i*settings->pairs->count] / EPS / EPS ;	
	}

	// compute integrals
	for (i=0;i<settings->pairs->count;i++)
	{
		t1[i] = fromb_t1(i, settings) ;		
		t2[i] = fromb_t2(i, settings) ;		
		t3[i] = fromb_t3(i, settings) ;		
		t4[i] = fromb_t4(i, settings) ;
		t5[i] = fromb_t5(i, settings) ;
	}
	
	// update fx	
	for (i=0;i<settings->pairs->count;i++)
		bfgs->fx += t1[i] ;
	// update gx 
	for (i=0;i<settings->number;i++)
		bfgs->gx[i] = 0 ;

	// for gx[0] --- ln\sigma		
	for (i=0;i<settings->pairs->count;i++)
		bfgs->gx[0] += (-t4[i]+t5[i]) ;
	bfgs->gx[0] = - bfgs->gx[0] / 2.0 * settings->noisevar ;
	// for gx[1] --- \theta_b^1		
	for (i=0;i<settings->pairs->count;i++)
		bfgs->gx[1] += (-t2[i]+t3[i]) ;	
	// for gx[2] --- ln\Delta^r
	for (j=2;j<settings->pairs->classes;j++)
	{
		for (i=0;i<settings->pairs->count;i++)
		{
			if (j==(settings->alpha+i)->pair->target)
				bfgs->gx[j] += t3[i] ;
			else if ((settings->alpha+i)->pair->target==settings->pairs->classes)
				bfgs->gx[j] -= t2[i] ;
			else if ((settings->alpha+i)->pair->target>j)
				bfgs->gx[j] += (-t2[i]+t3[i]) ;
		}
		bfgs->gx[j] = settings->intervals[j-1]*bfgs->gx[j] ;
	}

	// for kernel parameters
	// calculate for each settings->kappa Gaussian parameter
	if (1 == settings->ardon)
	{
		for (ii=0;ii<settings->pairs->dimen;ii++)
		{			
			// calculate dsigmam for each settings->kappa parameter
			for (i=0;i<settings->pairs->count;i++)
			{
				alpha = settings->alpha + i ;	
				//if ( I_One == alpha->setname )
				{
					for (j=0;j<=i;j++)
					{
						alphab = settings->alpha + j ;
						//if ( I_One == alphab->setname )
						{							
							if ( GAUSSIAN == settings->kernel )
							{
								dsigma[i*settings->pairs->count+j]=(alpha->kernels[j]-settings->kappa_m)*(-(alpha->pair->
									point[ii]-alphab->pair->point[ii])*(alpha->pair->point[ii]-alphab->pair->point[ii])/2.0) ;
							}
							else if ( LINEAR == settings->kernel )
							{
								dsigma[i*settings->pairs->count+j]=(alpha->pair->point[ii]*alphab->pair->point[ii]) ;
							}
							/*
							else if ( POLYNOMIAL == settings->kernel ) 
							{
								cbe = 0 ; 
								for ( iii = 0 ; iii < settings->pairs->dimen; iii++)
									cbe += settings->kappa[iii]*alpha->pair->point[iii]*alphab->pair->point[iii] ;
								if (cbe > 0)
									dsigma[i*settings->pairs->count+j]=alpha->kernels[j]*settings->p*alpha->pair->point[ii]
									*alphab->pair->point[ii]/(cbe) ;
								else
									dsigma[i*settings->pairs->count+j]=0 ;
							}*/
							else if ( USERDEFINED == settings->kernel )
							{
								cbe = 0 ; // temp storage
								if (0==settings->pairs->featuretype[ii])
									cbe-=(alpha->pair->point[ii]-alphab->pair->point[ii]) 
									*(alpha->pair->point[ii]-alphab->pair->point[ii])/2.0 ;
								else if (alpha->pair->point[ii]!=alphab->pair->point[ii])
									cbe-=1.0/2.0 ;
								dsigma[i*settings->pairs->count+j] = (alpha->kernels[j]-settings->kappa_m) * cbe ;
							}
							else
								printf ("Error in kernel type.\n") ;
							dsigma[j*settings->pairs->count+i] = dsigma[i*settings->pairs->count+j] ;
						}											
					}
				}
			}
			//brmul(adjsgmm, dsigmam, settings->pairs->count, settings->pairs->count, settings->pairs->count, sigdsig) ;
			temp = 0 ;
			for (i=0;i<settings->pairs->count;i++)
			{
				for (j=0;j<i;j++)
					temp += 2.0*settings->invcov[i*settings->pairs->count+j]*dsigma[i*settings->pairs->count+j] ;	
				temp += settings->invcov[i*settings->pairs->count+i]*dsigma[i*settings->pairs->count+i] ;	
			}
			bfgs->gx[ii+settings->pairs->classes] = -temp/2.0 ;
			// VC*VC*a'*dsigma*a/2
			for (i=0;i<settings->pairs->count;i++)
			{						
				for (j=i+1;j<settings->pairs->count;j++)
					bfgs->gx[ii+settings->pairs->classes] += (settings->alpha+i)->pair->weight*
					(settings->alpha+j)->pair->weight*dsigma[i*settings->pairs->count+j] ;		
				bfgs->gx[ii+settings->pairs->classes] += (settings->alpha+i)->pair->weight*
					(settings->alpha+i)->pair->weight*dsigma[i*settings->pairs->count+i]/2.0 ;				
			}
			bfgs->gx[ii+settings->pairs->classes] = bfgs->gx[ii+settings->pairs->classes] * settings->kappa[ii] ;
		}
	}
	else
	{
		for (i=0;i<settings->pairs->count;i++)
		{
			alpha = settings->alpha + i ;			
			//if ( I_One == alpha->setname )
			{
				for (j=0;j<=i;j++)
				{
					alphab = settings->alpha + j ;
					//if ( I_One == alphab->setname )
					{							
						if ( GAUSSIAN == settings->kernel )
						{
							cbe = 0 ; // temp storage
							for (ii=0;ii<settings->pairs->dimen;ii++)
								cbe-=(alpha->pair->point[ii]-alphab->pair->point[ii])
								*(alpha->pair->point[ii]-alphab->pair->point[ii]) ;
							dsigma[i*settings->pairs->count+j] = (alpha->kernels[j]-settings->kappa_m)*cbe/2.0 ;
						} 
						else if (LINEAR == settings->kernel )
						{
							cbe = 0 ; // temp storage
							for (ii=0;ii<settings->pairs->dimen;ii++)
								cbe += alpha->pair->point[ii]*alphab->pair->point[ii] ;
							dsigma[i*settings->pairs->count+j] = cbe ;

						}/*
						else if ( POLYNOMIAL == settings->kernel ) 
						{
							cbe = 0 ; 
							for (iii = 0 ; iii < settings->pairs->dimen; iii++)
								cbe += alpha->pair->point[iii]*alphab->pair->point[iii] ;
							if (cbe > 0)
								dsigma[i*settings->pairs->count+j] = alpha->kernels[j]*settings->p/settings->kappa_a ;
							else
								dsigma[i*settings->pairs->count+j] = 0 ;
						}*/
						else if ( USERDEFINED == settings->kernel )
						{
							cbe = 0 ; // temp storage
							for (ii=0;ii<settings->pairs->dimen;ii++)
							{
								if (0==settings->pairs->featuretype[ii])
									cbe-=(alpha->pair->point[ii]-alphab->pair->point[ii])
									*(alpha->pair->point[ii]-alphab->pair->point[ii]) ;
								else if (alpha->pair->point[ii]!=alphab->pair->point[ii])
									cbe-=1.0 ;
							}
							dsigma[i*settings->pairs->count+j] = (alpha->kernels[j]-settings->kappa_m) * cbe / 2.0 ;
						}
						else
							printf ("Error in kernel type.\n") ;
						dsigma[j*settings->pairs->count+i] = dsigma[i*settings->pairs->count+j] ;
					}
				}
			}
		}
		temp = 0 ;
		for (i=0;i<settings->pairs->count;i++)
		{
			for (j=0;j<i;j++)
				temp += 2.0*settings->invcov[i*settings->pairs->count+j]*dsigma[i*settings->pairs->count+j] ;	
			temp += settings->invcov[i*settings->pairs->count+i]*dsigma[i*settings->pairs->count+i] ;	
		}
		bfgs->gx[settings->pairs->classes] = - temp/2.0 ;
		// VC*VC*a'*dsigma*a/2
		for (i=0;i<settings->pairs->count;i++)
		{						
			for (j=0;j<i;j++)
				bfgs->gx[settings->pairs->classes] += (settings->alpha+i)->pair->weight*
				(settings->alpha+j)->pair->weight*dsigma[i*settings->pairs->count+j] ;		
			bfgs->gx[settings->pairs->classes] += (settings->alpha+i)->pair->weight*
				(settings->alpha+i)->pair->weight*dsigma[i*settings->pairs->count+i]/2.0 ;				
		}
		bfgs->gx[settings->pairs->classes] = bfgs->gx[settings->pairs->classes] * settings->kappa_a ;
	}

	bfgs->fx = - bfgs->fx ;
	for (i=0;i<settings->number;i++)
	{
		bfgs->gx[i] = - bfgs->gx[i] ;
		//printf("gardient %f --- x %f\n", bfgs->gx[i], bfgs->x[i]) ;
	}

	free(t5) ;
	free(t4) ;
	free(t3) ;
	free(t2) ;
	free(t1) ;
	free(dsigma) ;
	return 0 ;
}

int	Ordinal_EPEM_Training (Ordinal_Settings * settings) 
{
	unsigned int i ;

	if (NULL == settings)
		return 1 ;

	printf("Evidence Maximization on Lower Bound with EP ...\n") ;

	if ( 1 == bfgs_minfunc ( Ordinal_EPEM_Evaluate_FuncGrad, settings, settings->bfgs ) )
	{
		printf("fail to find local minimum.\n") ;
		return 1;
	}
	// save best funcgrad to settings 
	bfgs_ordinal_Settings( settings, settings->bfgs ) ;
	printf("Optimal Settings: \n") ;
	printf("Noise Variance: %.3f\n",settings->noisevar) ;
	printf("Kappa_O: %.3f\n",settings->kappa_o) ;
	printf("Thresholds:\n") ;
	for (i=1;i<settings->pairs->classes;i++)
		printf("%u ---  %.3f\n",i,settings->thresholds[i-1]) ;
	if (0==settings->ardon)		
		printf("Kernel: %.3f\n",settings->kappa_a) ;
	else
	{
		printf("Kernel:\n") ;
		for (i=0;i<settings->pairs->dimen;i++)
			printf("%u ---  %.3f:\n", i, settings->kappa[i]) ;
	}

	Ordinal_EP_Training (settings) ;

	return 0 ;
}

