#include <stdio.h>
#include <stdlib.h>
#include <math.h>
#include "ordinal.h"
// romberg for normcdf

double normpdf(double x)
{
	return exp(-0.5*x*x)/sqrt(2.0*PI) ;
}

double normcdf(double ca, double eb)
{ 
    int m,n,i,k;
	double eps ;
    double y[20],h,ep,p,x,s,q=0,a,b ;

	eps = EPS*EPS ;
	if (ca>eb)
	{	a=eb; b=ca;	}
	else
	{	a=ca; b=eb;	}

    h=b-a;
    y[0]=h*(normpdf(a)+normpdf(b))/2.0;
    m=1; n=1; ep=eps+1.0;
    while ((ep>=eps)&&(m<=19))
      { p=0.0;
        for (i=0;i<=n-1;i++)
          { x=a+(i+0.5)*h;
            p=p+normpdf(x);
          }
        p=(y[0]+h*p)/2.0;
        s=1.0;
        for (k=1;k<=m;k++)
          { s=4.0*s;
            q=(s*p-y[k-1])/(s-1.0);
            y[k-1]=p; p=q;
          }
        ep=fabs(q-y[m-1]);
        m=m+1; y[m-1]=q; n=n+n; h=h/2.0;
      }
	if (ca>eb)
    return(-q) ;
	else
    return(q) ;
}

double gaussian(double x, double mean, double sig)
{
	if (sig>0)
		return exp(-0.5*(x-mean)*(x-mean)/sig/sig)/sqrt(2.0*PI)/sig ;
	else
	{
		printf("negative variance in Gaussian.\n") ;
		return 0 ;
	}
}

double fromb_fft1(double b, double mean, double sig, int index, Ordinal_Settings * settings)
{
	double func ;
	double prob ;
	double z1, z2 ;	
	double phi1 = 1, phi2 = 0 ;
	Data_Node * node ;
	if (NULL == settings)
		return 0 ;
	func = gaussian(b,mean,sig) ;
	node = (settings->alpha+index)->pair ;
	if (2 <= node->target)
	{
		z2 = (settings->thresholds[node->target-2] - b)/sqrt(settings->noisevar) ;
		phi2 = 0.5 + 0.5 * lerrf(z2/sqrt(2)) ;
	}
	if (settings->pairs->classes-1 >= node->target) 			
	{
		z1 = (settings->thresholds[node->target-1] - b)/sqrt(settings->noisevar) ;
		phi1 = 0.5 + 0.5 * lerrf(z1/sqrt(2)) ;
	}
	prob = phi1 - phi2 ;
	if (prob<EPS*EPS)
		prob = EPS*EPS ;
	func = func*log(prob) ;
	return func ;
}

double fromb_t1(int index , Ordinal_Settings * settings)
{ 
    int m,n,i,k;
	double mean ;
	double sig ;
	double eps ;
    double y[20],h,ep,p,x,s,q=0;
	double a, b ;

	mean = (settings->alpha+index)->pair->postmean ;
	sig = sqrt((settings->alpha+index)->postcov[index]) ;
	eps = EPS ;

	a = mean-5.0*sig ;
	b = mean+5.0*sig ;

    h=b-a;
    y[0]=h*(fromb_fft1(a,mean,sig,index,settings)+fromb_fft1(b,mean,sig,index,settings))/2.0;
    m=1; n=1; ep=eps+1.0;
    while ((ep>=eps)&&(m<=19))
      { p=0.0;
        for (i=0;i<=n-1;i++)
          { x=a+(i+0.5)*h;
            p=p+fromb_fft1(x,mean,sig,index,settings);
          }
        p=(y[0]+h*p)/2.0;
        s=1.0;
        for (k=1;k<=m;k++)
          { s=4.0*s;
            q=(s*p-y[k-1])/(s-1.0);
            y[k-1]=p; p=q;
          }
        ep=fabs(q-y[m-1]);
        m=m+1; y[m-1]=q; n=n+n; h=h/2.0;
      }
    return(q);
}


double fromb_fft2(double b, double mean, double sig, int index, Ordinal_Settings * settings)
{
	double func ;
	double prob ;
	double z1, z2 ;	
	double phi1 = 1, phi2 = 0 ;
	Data_Node * node ;
	if (NULL == settings)
		return 0 ;
	func = gaussian(b,mean,sig) ;
	node = (settings->alpha+index)->pair ;
	if (2 <= node->target)
	{
		z2 = (settings->thresholds[node->target-2] - b)/sqrt(settings->noisevar) ;
		phi2 = 0.5 + 0.5 * lerrf(z2/sqrt(2)) ;
	}
	if (settings->pairs->classes-1 >= node->target) 			
	{
		z1 = (settings->thresholds[node->target-1] - b)/sqrt(settings->noisevar) ;
		phi1 = 0.5 + 0.5 * lerrf(z1/sqrt(2)) ;
	}
	prob = phi1 - phi2 ;
	if (prob<EPS*EPS)
		prob = EPS*EPS ;
	func = func / prob * gaussian(node->postmean,settings->thresholds[node->target-2],
		sqrt(settings->noisevar+(settings->alpha+index)->postcov[index])) ;
	return func ;
}

double fromb_t2(int index , Ordinal_Settings * settings)
  { 
    int m,n,i,k;
	double mean ;
	double sig ;
	double eps ;
    double y[20],h,ep,p,x,s,q=0;
	double a, b ;

	if (NULL == settings)
		return 0 ;
	if (1==(settings->alpha+index)->pair->target)
		return 0 ;

	mean = ((settings->alpha+index)->pair->postmean*settings->noisevar
		+(settings->alpha+index)->postcov[index]*settings->thresholds[(settings->alpha+index)->pair->target-2])
		/(settings->noisevar+(settings->alpha+index)->postcov[index]) ;
	sig = sqrt((settings->noisevar*(settings->alpha+index)->postcov[index])/(settings->noisevar+(settings->alpha+index)->postcov[index])) ;
	eps = EPS ;

	a = mean-5.0*sig ;
	b = mean+5.0*sig ;

    h=b-a;
    y[0]=h*(fromb_fft2(a,mean,sig,index,settings)+fromb_fft2(b,mean,sig,index,settings))/2.0;
    m=1; n=1; ep=eps+1.0;
    while ((ep>=eps)&&(m<=19))
      { p=0.0;
        for (i=0;i<=n-1;i++)
          { x=a+(i+0.5)*h;
            p=p+fromb_fft2(x,mean,sig,index,settings);
          }
        p=(y[0]+h*p)/2.0;
        s=1.0;
        for (k=1;k<=m;k++)
          { s=4.0*s;
            q=(s*p-y[k-1])/(s-1.0);
            y[k-1]=p; p=q;
          }
        ep=fabs(q-y[m-1]);
        m=m+1; y[m-1]=q; n=n+n; h=h/2.0;
      }
    return(q);
  }



double fromb_fft3(double b, double mean, double sig, int index, Ordinal_Settings * settings)
{
	double func ;
	double prob ;
	double z1, z2 ;	
	double phi1 = 1, phi2 = 0 ;
	Data_Node * node ;
	if (NULL == settings)
		return 0 ;
	func = gaussian(b,mean,sig) ;
	node = (settings->alpha+index)->pair ;
	if (2 <= node->target)
	{
		z2 = (settings->thresholds[node->target-2] - b)/sqrt(settings->noisevar) ;
		phi2 = 0.5 + 0.5 * lerrf(z2/sqrt(2)) ;
	}
	if (settings->pairs->classes-1 >= node->target) 			
	{
		z1 = (settings->thresholds[node->target-1] - b)/sqrt(settings->noisevar) ;
		phi1 = 0.5 + 0.5 * lerrf(z1/sqrt(2)) ;
	}
	prob = phi1 - phi2 ;
	if (prob<EPS*EPS)
		prob = EPS*EPS ;
	func = func / prob * gaussian(node->postmean,settings->thresholds[node->target-1],
		sqrt(settings->noisevar+(settings->alpha+index)->postcov[index])) ;
	return func ;
}

double fromb_t3(int index , Ordinal_Settings * settings)
  { 
    int m,n,i,k;
	double mean ;
	double sig ;
	double eps ;
    double y[20],h,ep,p,x,s,q=0;
	double a, b ;

	if (NULL == settings)
		return 0 ;
	if (settings->pairs->classes==(settings->alpha+index)->pair->target)
		return 0 ;

	mean = ((settings->alpha+index)->pair->postmean*settings->noisevar
		+(settings->alpha+index)->postcov[index]*settings->thresholds[(settings->alpha+index)->pair->target-1])
		/(settings->noisevar+(settings->alpha+index)->postcov[index]) ;
	sig = sqrt((settings->noisevar*(settings->alpha+index)->postcov[index])/(settings->noisevar+(settings->alpha+index)->postcov[index])) ;
	eps = EPS ;

	a = mean-5.0*sig ;
	b = mean+5.0*sig ;

    h=b-a;
    y[0]=h*(fromb_fft3(a,mean,sig,index,settings)+fromb_fft3(b,mean,sig,index,settings))/2.0;
    m=1; n=1; ep=eps+1.0;
    while ((ep>=eps)&&(m<=19))
      { p=0.0;
        for (i=0;i<=n-1;i++)
          { x=a+(i+0.5)*h;
            p=p+fromb_fft3(x,mean,sig,index,settings);
          }
        p=(y[0]+h*p)/2.0;
        s=1.0;
        for (k=1;k<=m;k++)
          { s=4.0*s;
            q=(s*p-y[k-1])/(s-1.0);
            y[k-1]=p; p=q;
          }
        ep=fabs(q-y[m-1]);
        m=m+1; y[m-1]=q; n=n+n; h=h/2.0;
      }
    return(q);
  }



double fromb_fft4(double b, double mean, double sig, int index, Ordinal_Settings * settings)
{
	double func ;
	double prob ;
	double z1, z2 ;	
	double phi1 = 1, phi2 = 0 ;
	Data_Node * node ;
	if (NULL == settings)
		return 0 ;
	func = gaussian(b,mean,sig) ;
	node = (settings->alpha+index)->pair ;
	if (2 <= node->target)
	{
		z2 = (settings->thresholds[node->target-2] - b)/sqrt(settings->noisevar) ;
		phi2 = 0.5 + 0.5 * lerrf(z2/sqrt(2)) ;
	}
	if (settings->pairs->classes-1 >= node->target) 			
	{
		z1 = (settings->thresholds[node->target-1] - b)/sqrt(settings->noisevar) ;
		phi1 = 0.5 + 0.5 * lerrf(z1/sqrt(2)) ;
	}
	prob = phi1 - phi2 ;
	if (prob<EPS*EPS)
		prob = EPS*EPS ;
	func = func / prob * gaussian(node->postmean,settings->thresholds[node->target-1],
		sqrt(settings->noisevar+(settings->alpha+index)->postcov[index])) *(settings->thresholds[node->target-1]-b) ;
	return func ;
}

double fromb_t4(int index , Ordinal_Settings * settings)
  { 
    int m,n,i,k;
	double mean ;
	double sig ;
	double eps ;
    double y[20],h,ep,p,x,s,q=0;
	double a, b ;

	if (NULL == settings)
		return 0 ;
	if (settings->pairs->classes==(settings->alpha+index)->pair->target)
		return 0 ;

	mean = ((settings->alpha+index)->pair->postmean*settings->noisevar
		+(settings->alpha+index)->postcov[index]*settings->thresholds[(settings->alpha+index)->pair->target-1])
		/(settings->noisevar+(settings->alpha+index)->postcov[index]) ;
	sig = sqrt((settings->noisevar*(settings->alpha+index)->postcov[index])/(settings->noisevar+(settings->alpha+index)->postcov[index])) ;
	eps = EPS ;

	a = mean-5.0*sig ;
	b = mean+5.0*sig ;

    h=b-a;
    y[0]=h*(fromb_fft4(a,mean,sig,index,settings)+fromb_fft4(b,mean,sig,index,settings))/2.0;
    m=1; n=1; ep=eps+1.0;
    while ((ep>=eps)&&(m<=19))
      { p=0.0;
        for (i=0;i<=n-1;i++)
          { x=a+(i+0.5)*h;
            p=p+fromb_fft4(x,mean,sig,index,settings);
          }
        p=(y[0]+h*p)/2.0;
        s=1.0;
        for (k=1;k<=m;k++)
          { s=4.0*s;
            q=(s*p-y[k-1])/(s-1.0);
            y[k-1]=p; p=q;
          }
        ep=fabs(q-y[m-1]);
        m=m+1; y[m-1]=q; n=n+n; h=h/2.0;
      }
    return(q);
  }


double fromb_fft5(double b, double mean, double sig, int index, Ordinal_Settings * settings)
{
	double func ;
	double prob ;
	double z1, z2 ;	
	double phi1 = 1, phi2 = 0 ;
	Data_Node * node ;
	if (NULL == settings)
		return 0 ;
	func = gaussian(b,mean,sig) ;
	node = (settings->alpha+index)->pair ;
	if (2 <= node->target)
	{
		z2 = (settings->thresholds[node->target-2] - b)/sqrt(settings->noisevar) ;
		phi2 = 0.5 + 0.5 * lerrf(z2/sqrt(2)) ;
	}
	if (settings->pairs->classes-1 >= node->target) 			
	{
		z1 = (settings->thresholds[node->target-1] - b)/sqrt(settings->noisevar) ;
		phi1 = 0.5 + 0.5 * lerrf(z1/sqrt(2)) ;
	}
	prob = phi1 - phi2 ;
	if (prob<EPS*EPS)
		prob = EPS*EPS ;
	func = func / prob * gaussian(node->postmean,settings->thresholds[node->target-2],
		sqrt(settings->noisevar+(settings->alpha+index)->postcov[index])) * (settings->thresholds[node->target-2]-b) ;
	return func ;
}

double fromb_t5(int index , Ordinal_Settings * settings)
  { 
    int m,n,i,k;
	double mean ;
	double sig ;
	double eps ;
    double y[20],h,ep,p,x,s,q=0;
	double a, b ;

	if (NULL == settings)
		return 0 ;
	if (1==(settings->alpha+index)->pair->target)
		return 0 ;

	mean = ((settings->alpha+index)->pair->postmean*settings->noisevar
		+(settings->alpha+index)->postcov[index]*settings->thresholds[(settings->alpha+index)->pair->target-2])
		/(settings->noisevar+(settings->alpha+index)->postcov[index]) ;
	sig = sqrt((settings->noisevar*(settings->alpha+index)->postcov[index])/(settings->noisevar+(settings->alpha+index)->postcov[index])) ;
	eps = EPS ;

	a = mean-5.0*sig ;
	b = mean+5.0*sig ;

    h=b-a;
    y[0]=h*(fromb_fft5(a,mean,sig,index,settings)+fromb_fft5(b,mean,sig,index,settings))/2.0;
    m=1; n=1; ep=eps+1.0;
    while ((ep>=eps)&&(m<=19))
      { p=0.0;
        for (i=0;i<=n-1;i++)
          { x=a+(i+0.5)*h;
            p=p+fromb_fft5(x,mean,sig,index,settings);
          }
        p=(y[0]+h*p)/2.0;
        s=1.0;
        for (k=1;k<=m;k++)
          { s=4.0*s;
            q=(s*p-y[k-1])/(s-1.0);
            y[k-1]=p; p=q;
          }
        ep=fabs(q-y[m-1]);
        m=m+1; y[m-1]=q; n=n+n; h=h/2.0;
      }
    return(q);
  }


//end of fromb.c

