#include <stdio.h>
#include <stdlib.h>
#include <string.h>
#include <time.h>
#include <math.h>
#include "ordinal.h"
//#include "bit_matrix.h"

extern double lerrf(double) ;
extern double fromb(double mean, double sig, double eps) ;

int Ordinal_Prediction (Ordinal_Settings * settings) 
{
	Data_List testlist ;
	Data_Node * node ;
	Data_List * trainlist ;

	unsigned int i, k ;
	char buf[2048] ;
	double * kcvx ;
	double * prob ;
	double covx ;
	double z1, z2 ;
	unsigned int index = 0 ;
	double phi1, phi2 ;
	double guess ;
	FILE * fid ;
	FILE * fid_guess ; 
	unsigned int error = 0, abserr = 0 ;

	//FILE * guess_file = NULL ; // filename.guess
	//FILE * cfd_file = NULL ; // filename.conf
	//FILE * func_file = NULL ; // filename.func
	//FILE * prob_file = NULL ; // filename.prob
	//FILE * log_file = NULL ; // filename.log

	if (NULL == settings)
		return 1 ;
	
	trainlist = settings->pairs ;

	// loading test data 
	Create_Data_List (&testlist) ;
	ordinal_Loadfile ( &testlist, settings->testfile, trainlist->dimen ) ; 
	
	// compute predictions
	node = testlist.front ;
	while (NULL != node)
	{
		// for each new point
		// transform input point
		if (1 == trainlist->normalized_input)
		{
			for (k=0;k<testlist.dimen;k++)
			{ 
				if ( 0 != trainlist->x_devi[k] )
					node->point[k] = (node->point[k]-trainlist->x_mean[k])/(trainlist->x_devi[k]) ;
				else
					node->point[k] = 0 ;
			}
		}
		node->pred_func = 0 ;
		// cache kcx 
		kcvx = (double*)malloc(trainlist->count*sizeof(double)) ;
		for (i=0;i<trainlist->count;i++)
		{
			kcvx[i] = Calc_Covfun((settings->alpha+i)->pair->point,node->point,settings) ;
			node->pred_func += (settings->alpha+i)->pair->weight*kcvx[i];
		}
		covx = Calc_Covfun(node->point,node->point,settings) ; ;
		// compute sigmax
		for (i=0;i<trainlist->count;i++)
		{
			for (k=0;k<i;k++)
				covx -= 2.0*kcvx[i]*kcvx[k]*settings->invcov[i*trainlist->count+k] ;
			covx -= kcvx[i]*kcvx[i]*settings->invcov[i*trainlist->count+i] ;
		}
		if (covx>0)
			node->pred_var = covx ;
		else
			printf("Warning: %u sample with a negative variance.",index+1) ;
		free(kcvx) ;
		node=node->next ;
		index += 1 ;
	}

	// save the function values
	sprintf(buf,"%s.func",settings->testfile) ;
	fid = fopen(buf,"w+t") ;
	if (NULL != fid)
	{
		node = testlist.front ;
		while (NULL != node)
		{
			// for each new 
			fprintf(fid, "%f\n",node->pred_func) ;
			node=node->next ;
		}
		fclose(fid) ;
	}	
	// save the variance values
	sprintf(buf,"%s.conf",settings->testfile) ;
	fid = fopen(buf,"w+t") ;
	if (NULL != fid)
	{
		node = testlist.front ;
		while (NULL != node)
		{
			// for each new 
			fprintf(fid, "%f\n",node->pred_var) ;
			node=node->next ;
		}
		fclose(fid) ;
	}
	// compute predictions
	sprintf(buf,"%s.prob",settings->testfile) ;
	fid = fopen(buf,"w+t") ;
	sprintf(buf,"%s.guess",settings->testfile) ;
	fid_guess = fopen(buf,"w+t") ;
	prob = (double*) malloc(trainlist->classes*sizeof(double)) ;
	node = testlist.front ;	
	while (NULL != node)
	{
		// compute predictive probability
		guess = 0 ;
		for (i=1;i<=trainlist->classes;i++)
		{			
			phi2 = 0 ;
			phi1 = 1 ;
			if (2 <= i)
			{
				z2 = (settings->thresholds[i-2] - node->pred_func)/sqrt(node->pred_var+settings->noisevar) ;
				phi2 = 0.5 + 0.5 * lerrf(z2) ;
			}
			if (trainlist->classes-1 >= i) 	
			{
				z1 = (settings->thresholds[i-1] - node->pred_func)/sqrt(node->pred_var+settings->noisevar) ;
				phi1 = 0.5 + 0.5 * lerrf(z1) ;
			}
			prob[i-1] = phi1 - phi2 ;  
			if (NULL != fid)
				fprintf(fid,"%f ",prob[i-1]) ;
			if (prob[i-1]>guess)
			{
				guess = prob[i-1] ;
				node->pred_label = i ;
			}
		}
		if (node->pred_label != node->target)
		{
			error += 1 ;
			abserr += abs(node->pred_label-node->target) ;
		}

		if (NULL != fid)
			fprintf(fid,"\n") ;
		if (NULL != fid_guess)
			fprintf(fid_guess,"%u\n",node->pred_label) ;
		node=node->next ;
	}
	free(prob) ;
	if (NULL != fid)
		fclose(fid) ;
	if (NULL != fid_guess)
		fclose(fid_guess) ;

	if (testlist.deviation > 0)
		printf("Testing error number %u with absolute error %u.\n", error, abserr) ;
	else
		printf("Targets are not available in testing data file %s.\n", testlist.filename) ;
#ifdef  _ORDINAL_LP
	fid = fopen ("gpordinal_lap_batch.log","a+t") ;
#endif
#ifdef  _ORDINAL_EP
	fid = fopen ("gpordinal_epem_batch.log","a+t") ;
#endif
	if (NULL != fid)
	{
		fprintf(fid,"%u %u %f\n", error, abserr, (double)abserr/(double)(testlist.count)) ;
		fclose(fid) ;
	}

	// save log files
	Clear_Data_List (&testlist) ;
	return 0 ;
}

int Dumping_Ordinal_Settings (Ordinal_Settings * settings)
{
	FILE * fid ;
	char buf[2048] ;
	unsigned int i ;

	if (NULL == settings)
		return 1 ;

	sprintf(buf,"%s.log",settings->pairs->filename) ;
	fid = fopen (buf,"w+t") ;
	if (NULL != fid)
	{
		// add testing information
		sprintf(buf, "final settings\n") ;
		fwrite (buf, sizeof(char), strchr(buf,'\0')-buf, fid) ;
		fprintf(fid,"kappa_o = %f\n",settings->kappa_o) ;
		if (0 == settings->ardon)
			fprintf(fid,"kappa = %f\n",settings->kappa_a) ;
		else 
		{
			for (i=0;i<settings->pairs->dimen;i++)
				fprintf(fid,"kappa%d = %f\n", i+1, settings->kappa[i]) ;				
		}
		fprintf(fid,"kappa_m = %f\n", settings->kappa_m) ;
		fprintf(fid,"noise variance = %f\n", settings->noisevar) ;
		for (i=0;i<settings->pairs->classes-1;i++)
			fprintf(fid,"threshold%d = %f\n", i+1, settings->thresholds[i]) ;
		fclose(fid) ;
	}
	return 0 ;
}

// end of gpor_predict.cpp

