// main function 

// Chu Wei (C) Copyright 2004 IPAM & GATSBY

#include <stdio.h>
#include <stdlib.h>
#include <math.h>
#include <string.h>
#include <float.h>
#include "ordinal.h"
#define VERSION (1)
#define TRIALS (6)
#define TOPGENES (50)
#define _WINNER_TAKEALL
#define min(a,b) ((a) <= (b) ? (a) : (b))
#define max(a,b) ((a) >= (b) ? (a) : (b))


int main( int argc, char * argv[] )
{
	Data_List rawdatalist ;
	Data_Node * node ;
	Ordinal_Settings * settings ;
	double linear = 1 ;
	double kappa_o = 0 ;
	double kappa = 0 ;
	double kappa_m = 0 ;
	double regular = -1 ;
	unsigned int sz = 0 ;
	unsigned int index = 0 ;
	double parameter = 0 ;
	char buf[LENGTH]="" ;
	char filename[2048]="" ;
	double * results ;
	//double points[TRIALS]={0.05,0.025,0.01,0.005,0.0025,0.001,0.00075,0.0005,0.00025,0.0001} ;
	double points[TRIALS]={0.0001,0.001,0.01,0.1,1,10} ;
	double yards[TRIALS] ;
	int i ;
	FILE * fid ;
	float * ard ;
	unsigned long * rank ;
	double sum ;

	printf("\n Gaussian Processes for Gene Expression Analysis - version 1.%d\n - Chu Wei Copyright(C) 2004 at Gatsby Unit\n\n", VERSION) ;

	if ( 1 == argc )
	{
		// display help 
		printf(" Usage    : gpor [-K k] [-M m] [-O o][-R r] datafile\n") ;	
		printf(" - datafile:\t the data file of ordinal regression\n") ;
		printf(" - K k:\t set initial Kappa at k (default 1)\n") ;
		printf(" - M m:\t set initial kappa_M at m (default 0)\n") ;
		printf(" - O o:\t set initial kappa_O at o (default 1)\n") ;
		//printf(" - L  :\t set Linear kernel (default Gaussian)\n") ;
		printf(" - R r:\t set Regularization factor at r (default 1)\n\n") ;
		return 0 ;
	}
	else
	{
		sprintf(filename,"%s",argv[--argc]) ;
		if (argc>1)
			printf("Options:\n") ;
		do
		{
			strcpy(buf, argv[--argc]) ;
			sz = strlen(buf) ;
			//printf ("%s  %d\n", buf, sz) ;
			if ( '-' == buf[0] )
			{			
				for (index = 1 ; index < sz ; index++)
				{
					switch (buf[index])
					{
					case '-' :
						break ;
					case 'K' :						
						if (parameter > 0)
						{ 
							if (index + 1 == sz)
							{
								printf("  - initialize Kappa at the value %f.\n", parameter) ;
								kappa = parameter ;
								parameter = 0;
							}
						}
						break ;
					case 'O' :						
						if (parameter > 0)
						{ 
							if (index + 1 == sz)
							{
								printf("  - initialize kappa_O at the value %f.\n", parameter) ;
								kappa_o = parameter ;
								parameter = 0;
							}
						}
						break ;
					case 'M' :
						if (parameter > 0)
						{ 
							if (index + 1 == sz)
							{
								printf("  - initialize kappa_M at the value %f.\n", parameter) ;
								kappa_m = parameter ;
								parameter = 0;
							}
						}
						break ;
					case 'L' :
						printf("  - use LINEAR kernel.\n") ;
						linear = 0 ;
						break ;
					case 'R' :						
						if (parameter >= 0)
						{ 
							if (index + 1 == sz)
							{
								printf("  - initialize Regularization factor at the value %f.\n", parameter) ;
								regular = parameter ;
								parameter = 0;
							}
						}
						break ;
					default :
						printf("  - %c is invalid.\n", buf[index]) ;
						break ;
					}
				}
			}
			else
				parameter = atof(buf) ;
		}
		while ( argc > 1 ) ;
		printf("\n") ;
	}

	Create_Data_List (&rawdatalist) ;
	if ( 1 == ordinal_Loadfile (&rawdatalist, filename, 0) )
	{
		printf("fail to load data from %s.\n\n",filename) ;
		exit(1) ;
	}

	// try several initial points
	//for (index=0; index<TRIALS; index++)
	//{
	//	points[index]=1.0/(double)(rawdatalist.dimen)*pow(2.0,(double)((double)index*2.0-TRIALS)) ;
	//}
	// malloc 
	results = (double *) malloc(rawdatalist.dimen*TRIALS*sizeof(double)) ;

	for (index=0; index<TRIALS; index++)
	{

//##LOOPING
        points[index] = points[index]/(sqrt((double)rawdatalist.dimen)) ;
	kappa = points[index] ; // starting point

	settings = Create_Ordinal_Settings(&rawdatalist) ;
	settings->ardon = linear ; // use ARD
	linear = 1 ; // use linear
	if (NULL == settings)
	{
		printf("fail to generate ordinal settings.\n") ;
		exit(1) ;
	}
	else 
	{		
		if (kappa>0)
		{
				if (0==settings->ardon)		
					settings->kappa_a = kappa ;
				else
				{				
					for (i=0;i<(int)settings->pairs->dimen;i++)
						settings->kappa[i] = kappa ;
				}
		}
		if (kappa_m>0)
			settings->kappa_m = kappa_m ;
		if (kappa_o>0)
			settings->kappa_o = kappa_o ;
		if (regular>=0)
			settings->regular = regular ;
		if (linear>0.5)
			settings->kernel = LINEAR ;
	}

	settings->bfgs = Create_bfgs_FuncGrad(settings->number) ;
	if (NULL == settings->bfgs)
	{
		printf("fail to create BFGSs.\n") ;
		exit(1) ;
	}
	ordinal_bfgs_Settings(settings,settings->bfgs) ;

#ifdef _ORDINAL_EP
	Ordinal_EPEM_Training (settings) ;
#endif
#ifdef _ORDINAL_LP
    Ordinal_LAPLACE_Training (settings) ;
#endif

	//Ordinal_Prediction (settings) ;
	
	// SAVING RESULTS
	yards[index] = settings->bfgs->fx ;
	sum = 0 ;
	for (i=0; i<(int)settings->pairs->dimen; i++)
		sum += settings->kappa[i] ;
	for (i=0; i<(int)settings->pairs->dimen; i++)
	{
		if (sum>0)
			results[i+index*settings->pairs->dimen] = settings->kappa[i]/sum ;
		else
			results[i+index*settings->pairs->dimen] = 0 ;
	}
 	
	fid = fopen("logfile","a+t") ;
        for (i=0;i<(int)rawdatalist.dimen;i++)
                fprintf(fid,"%f ",results[i+index*settings->pairs->dimen]) ;
        fprintf(fid,"%f \n",yards[index]) ;
        fclose(fid);
      	
	Dumping_Ordinal_Settings(settings) ;
        Clear_Ordinal_Settings(settings) ;

//##LOOPING
	}

#ifdef _WINNER_TAKEALL
	// voting the best
	linear = yards[0] ; 
	i=0;
	for (index=1;index<TRIALS;index++)
	{
		if (yards[index]<linear)
		{
			linear = yards[index] ;
			i = index ;
		}
	}
	printf("The %d-th results starting from kappa=%f is selected.\n", i+1, points[i]) ;

	// sorting the genes

	ard = (float *) malloc((rawdatalist.dimen+1)*sizeof(float)) ;
	rank = (unsigned long *) malloc((rawdatalist.dimen+1)*sizeof(unsigned long)) ;
	for (index=0;index<(int)rawdatalist.dimen;index++)
	{
		ard[index+1] = (float)results[index+i*rawdatalist.dimen] ;
	}
	indexx(rawdatalist.dimen,ard,rank) ;
	for (index=0;index<(int)rawdatalist.dimen;index++)
	{
		ard[index+1] = (float)results[index+i*rawdatalist.dimen] ;
	}
#endif
#ifndef _WINNER_TAKEALL
	ard = (float *) calloc((rawdatalist.dimen+1),sizeof(float)) ;
	rank = (unsigned long *) malloc((rawdatalist.dimen+1)*sizeof(unsigned long)) ;
	sum = 0 ;
	for (i=0; i<TRIALS; i++)
	{	
		if (fabs(yards[i])<1.0/EPS)
			sum += exp(-yards[i]+yards[0]) ;
	}

	for (i=0;i<TRIALS;i++)
	{
		for (index=0;index<(int)rawdatalist.dimen;index++)
		{
			if (fabs(yards[i])<1.0/EPS)
				ard[index+1] += (float)(results[index+i*rawdatalist.dimen]*exp(-yards[i]+yards[0])/sum) ;
		}
	}
        sprintf(buf, "%s.ardvalue", filename) ;
        fid = fopen(buf,"w+t") ;
        for (index=0;index<rawdatalist.dimen;index++)            
                fprintf(fid, "%f\n", ard[index+1]) ;
        fclose(fid) ;           

	indexx(rawdatalist.dimen,ard,rank) ;
	free(ard);
        ard = (float *) calloc((rawdatalist.dimen+1),sizeof(float)) ;
	for (i=0;i<TRIALS;i++)
        {
                for (index=0;index<(int)rawdatalist.dimen;index++)
                {
                        if (fabs(yards[i])<1.0/EPS)
                                ard[index+1] += (float)(results[index+i*rawdatalist.dimen]*exp(-yards[i])/sum) ;
                }
        }
#endif
	
	// generating files
	for (index=1;index<=min(TOPGENES,rawdatalist.dimen);index++)
	{
		sprintf(buf, "%s.%u", filename, index) ;
		fid = fopen(buf,"w+t") ;		
		node = rawdatalist.front ;
		while (NULL != node)
		{

		for (i=1;i<=(int)index;i++) //features
		{
			fprintf(fid, "%f ", node->point[ rank[rawdatalist.dimen+1-(unsigned long)i]-1 ]) ;
		}
		fprintf(fid, "%u\n", node->target) ;
		node = node->next ;
		
		}
		fclose(fid) ;
		
		sprintf(buf, "%s.%u.feat", filename, index) ;
		fid = fopen(buf,"w+t") ;
		for (i=1;i<=(int)index;i++)		
			fprintf(fid, "%d\n", (int)rank[rawdatalist.dimen+1-(unsigned long)i]) ;
		fclose(fid) ;		
	}
	// save the feature ranks
	sprintf(buf, "%s.feat", filename) ;
	fid = fopen(buf,"w+t") ;
	for (index=1;index<=min(TOPGENES,rawdatalist.dimen);index++)		
		fprintf(fid, "%d %f\n", (int)rank[rawdatalist.dimen+1-index],ard[ rank[rawdatalist.dimen+1-index] ]) ;
	fclose(fid) ;		

	free(rank) ;
	free(ard) ;
	free(results) ;
	Clear_Data_List (&rawdatalist) ;
	return 0 ;
}


