#include <stdio.h>
#include <stdlib.h>
#include <math.h>
#include "ordinal.h"
#include "neal_matrix.h"

int Ordinal_LAPLACE_Evaluate_FuncGrad ( void * pointer, bfgs_FuncGrad * bfgs ) 
{
	Ordinal_Settings * settings ;
	Alphas * alpha ;
	Alphas * alphab ;
	double * t1 ;
	double * t2 ;
	double * w1 ;
	double * w2 ;
	double * g1 ;
	double * g2 ; 
	double * v1 ;
	double * v2 ; 
	double * q1 ;
	double * q2 ; 
	double * diag ;
	double * cache ;
	double * dsigma ;
	double * dmat ;
	double temp ;
	double cbe ;
	unsigned int i ;
	unsigned int j ;
	unsigned int ii, iii ; // settings->kappa	
	double z1, z2, phi1, phi2, n1, n2, dphi ;
	double sigma ;
	double di,dj,dij ;
	unsigned int dindex ;



	if (NULL == pointer || NULL == bfgs)
		return 1 ;
	settings = pointer ;

	bfgs_ordinal_Settings( settings,settings->bfgs ) ;

	// call EP routine to find posterior distribution 
	if (1 == Ordinal_MAP_Training (settings) )
	{
		printf("MAP failed to get equilibrium.\n") ;
		return 1 ;
	}

	// evaluate functional and gradient
	if (NULL != settings->invcov)
		free(settings->invcov) ;

	settings->invcov = (double*)malloc(settings->pairs->count*settings->pairs->count*sizeof(double)) ;
	dsigma = (double*)malloc(settings->pairs->count*settings->pairs->count*sizeof(double)) ;
	dmat = (double*)malloc(settings->pairs->count*settings->pairs->count*sizeof(double)) ;

	t1 = (double*)malloc(settings->pairs->count*sizeof(double)) ;
	t2 = (double*)malloc(settings->pairs->count*sizeof(double)) ;	
	w1 = (double*)malloc(settings->pairs->count*sizeof(double)) ;
	w2 = (double*)malloc(settings->pairs->count*sizeof(double)) ;	
	g1 = (double*)malloc(settings->pairs->count*sizeof(double)) ;
	g2 = (double*)malloc(settings->pairs->count*sizeof(double)) ;	
	v1 = (double*)malloc(settings->pairs->count*sizeof(double)) ;
	v2 = (double*)malloc(settings->pairs->count*sizeof(double)) ;	
	q1 = (double*)malloc(settings->pairs->count*sizeof(double)) ;
	q2 = (double*)malloc(settings->pairs->count*sizeof(double)) ;
	diag = (double*)malloc(settings->pairs->count*sizeof(double)) ;	
	cache = (double*)malloc(settings->pairs->count*sizeof(double)) ;


	sigma = sqrt(settings->noisevar) ;

	bfgs->fx = 0 ;
	// initialize epinvvar
	for (i=0;i<settings->pairs->count;i++)
	{
		n1 = 0 ;
		n2 = 0 ;
		z1 = 0 ;
		z2 = 0 ;
		phi1 = 1 ;
		phi2 = 0 ;

		if (1 == (settings->alpha+i)->pair->target)
		{
			z1 = (settings->thresholds[(settings->alpha+i)->pair->target-1]-(settings->alpha+i)->pair->epmean)/sigma ;
			n1 = normpdf(z1) ;
			phi1 = 0.5 + normcdf(0,z1) ;
			dphi = phi1-phi2 ;

		}
		else if (settings->pairs->classes == (settings->alpha+i)->pair->target)
		{
			z2 = (settings->thresholds[(settings->alpha+i)->pair->target-2]-(settings->alpha+i)->pair->epmean)/sigma ;
			n2 = normpdf(z2) ;
			phi2 = 0.5 + normcdf(0,z2) ;
			dphi = phi1-phi2 ;
		}
		else
		{
			z1 = (settings->thresholds[(settings->alpha+i)->pair->target-1]-(settings->alpha+i)->pair->epmean)/sigma ;
			z2 = (settings->thresholds[(settings->alpha+i)->pair->target-2]-(settings->alpha+i)->pair->epmean)/sigma ;			
			n1 = normpdf(z1) ;
			n2 = normpdf(z2) ;
			dphi = normcdf(z2,z1);
		}

		if (dphi<EPS)
		{			
			if (n1-n2<EPS)
				bfgs->fx -= log(EPS) ;
			else
				bfgs->fx -= log(EPS) ;
			w1[i] = 0 ;
			w2[i] = 0 ;
			g1[i] = 0 ;
			g2[i] = 0 ;
			v1[i] = 0 ;
			v2[i] = 0 ;
			q1[i] = 0 ;
			q2[i] = 0 ;
			if (1 == (settings->alpha+i)->pair->target)
			{
				(settings->alpha+i)->pair->weight = -z1/sigma ;
				(settings->alpha+i)->pair->epinvvar = 1/sigma/sigma ;
				w1[i] = -z1/sigma ;
				g1[i] = -z1*z1/sigma ;
				v1[i] = -z1*z1*z1/sigma ;
				q1[i] = -z1*z1*z1*z1/sigma ;
			}
			else if (settings->pairs->classes == (settings->alpha+i)->pair->target)
			{
				(settings->alpha+i)->pair->weight = z2/sigma ;
				(settings->alpha+i)->pair->epinvvar = 1/sigma/sigma ;				
				w2[i] = -z2/sigma ;
				g2[i] = -z2*z2/sigma ;
				v2[i] = -z2*z2*z2/sigma ;
				q2[i] = -z2*z2*z2*z2/sigma ;
			}
			else
			{
				if ((n1-n2)>=0)
				{
					(settings->alpha+i)->pair->weight = -(z1*exp(-0.5*z1*z1+0.5*z2*z2)-z2)/(exp(-0.5*z1*z1+0.5*z2*z2)-1.0)/sigma ;
					(settings->alpha+i)->pair->epinvvar = 1/sigma/sigma 
						+ (settings->alpha+i)->pair->weight*(settings->alpha+i)->pair->weight
						- (z1*z1*exp(-0.5*z1*z1+0.5*z2*z2)-z2*z2)/(exp(-0.5*z1*z1+0.5*z2*z2)-1.0)/sigma/sigma ;
					w1[i] = -z1*n1/(n1-n2) ;
					g1[i] = -z1*z1*n1/(n1-n2) ;
					v1[i] = -z1*z1*z1*n1/(n1-n2) ;
					q1[i] = -z1*z1*z1*z1*n1/(n1-n2) ;
					w2[i] = -z2*n2/(n1-n2) ;
					g2[i] = -z2*z2*n2/(n1-n2) ;
					v2[i] = -z2*z2*z2*n2/(n1-n2) ;
					q2[i] = -z2*z2*z2*z2*n2/(n1-n2) ;
				}
				else
				{
					//printf("n1 < n2 ;\n") ;
					(settings->alpha+i)->pair->epmean = 0 ;
					(settings->alpha+i)->pair->postmean = 0 ;
					(settings->alpha+i)->pair->weight = 0 ;
					(settings->alpha+i)->pair->epinvvar = 1/sigma/sigma ;
				}
			}
		}
		else
		{
			bfgs->fx -= log(dphi) ;
			// compute the moments
			w1[i] = n1/(dphi) ;
			w2[i] = n2/(dphi) ;
			g1[i] = z1*n1/(dphi) ;
			g2[i] = z2*n2/(dphi) ;
			v1[i] = z1*z1*n1/(dphi) ;
			v2[i] = z2*z2*n2/(dphi) ;
			q1[i] = z1*z1*z1*n1/(dphi) ;
			q2[i] = z2*z2*z2*n2/(dphi) ;			
			(settings->alpha+i)->pair->weight = (n1-n2)/(dphi)/sigma ;
			(settings->alpha+i)->pair->epinvvar = (settings->alpha+i)->pair->weight*(settings->alpha+i)->pair->weight 
				+ (g1[i]-g2[i])/settings->noisevar ;
		}
		(settings->alpha+i)->pair->weight = - (settings->alpha+i)->pair->weight ;
		if ((settings->alpha+i)->pair->epinvvar>1.0/sigma/sigma)
		{
			//printf("error : %u, %f\n", i, (settings->alpha+i)->pair->epinvvar) ;
			(settings->alpha+i)->pair->epinvvar = 1.0-EPS*EPS ;
		}
		if ((settings->alpha+i)->pair->epinvvar<0)
		{
			//printf("error : %u, %f\n", i, (settings->alpha+i)->pair->epinvvar) ;
			(settings->alpha+i)->pair->epinvvar = EPS*EPS ;
		}
	}
	
	for (i=0;i<settings->pairs->count;i++)
	{	
		for (j=0;j<i;j++)
		{
			settings->invcov[i*settings->pairs->count+j] = (settings->alpha+i)->kernels[j] ;
			settings->invcov[j*settings->pairs->count+i] = settings->invcov[i*settings->pairs->count+j] ;
			bfgs->fx += (settings->alpha+i)->kernels[j]
				*(settings->alpha+j)->pair->weight*(settings->alpha+i)->pair->weight ;
		}
		settings->invcov[i*settings->pairs->count+i] = (settings->alpha+i)->kernels[i] ;
		bfgs->fx += 0.5 * (settings->alpha+i)->kernels[i]
			*(settings->alpha+i)->pair->weight*(settings->alpha+i)->pair->weight ;
		settings->invcov[i*settings->pairs->count+i] += 1.0/(settings->alpha+i)->pair->epinvvar ;
	}
	if (0 == cholesky(settings->invcov,settings->pairs->count,&temp))
		printf("Fatal Error: Sigma+1/Pi is not positive definite.\n") ;

	for (i=0;i<settings->pairs->count;i++)
	{
		bfgs->fx += log(settings->invcov[i+i*settings->pairs->count]) ;
		if ( (settings->alpha+i)->pair->epinvvar > 0 )
			bfgs->fx += 0.5*log((settings->alpha+i)->pair->epinvvar);
		else
			bfgs->fx += 0.5*log(EPS*EPS);
	}
	// end of fx

	if (0 == inverse_from_cholesky(settings->invcov, t1, t2, settings->pairs->count) )		
		printf("Fatal Error: Sigma+1/Pi is not positive definite.\n") ;

	// compute a diagnal 	
	for (i=0;i<settings->pairs->count;i++)	
	{
		for (j=0;j<settings->pairs->count;j++)
		{
			dsigma[i*settings->pairs->count+j] = 0 ;
			for (ii=0;ii<settings->pairs->count;ii++)
				dsigma[i*settings->pairs->count+j] += settings->invcov[i*settings->pairs->count+ii]
					*Calculate_Covfun(settings->alpha+ii,settings->alpha+j,settings) ;			
		}
		diag[i] = dsigma[i*settings->pairs->count+i] / (settings->alpha+i)->pair->epinvvar ;
	}

	//  partial lambda / partial theta_b = - partial lambda / partial f  (*SIGMA)
	for (i=0;i<settings->pairs->count;i++)
		t2[i] = ( (w1[i]-w2[i])-3.0*(w1[i]-w2[i])*(g1[i]-g2[i])
		-2.0*(w1[i]-w2[i])*(w1[i]-w2[i])*(w1[i]-w2[i])
		-(v1[i]-v2[i]) ) / settings->noisevar ;
	
	// update gx 
	for (i=0;i<settings->number;i++)
		bfgs->gx[i] = 0 ;

	// for gx[0] --- ln\sigma
	
	for (i=0;i<settings->pairs->count;i++)
		cache[i] = ((w1[i]-w2[i])*(g1[i]-g2[i])-(w1[i]-w2[i])+(v1[i]-v2[i]))/settings->noisevar ;

	// prepare D f / D delta_l
	for (ii=0;ii<settings->pairs->count;ii++)
	{
		t1[ii] = 0 ;
		for (iii=0;iii<settings->pairs->count;iii++)
			t1[ii] += dsigma[ii*settings->pairs->count+iii] * cache[iii] ;
		t1[ii] = - t1[ii] / (settings->alpha+ii)->pair->epinvvar ;
	}		
	
	for (i=0;i<settings->pairs->count;i++)
	{
		temp = -2.0*(settings->alpha+i)->pair->epinvvar + (2.0*(w1[i]-w2[i])*
		(v1[i]-v2[i])+2.0*(w1[i]-w2[i])*(w1[i]-w2[i])*(g1[i]-g2[i])-(g1[i]-g2[i])
		+(g1[i]-g2[i])*(g1[i]-g2[i])+q1[i]-q2[i])/settings->noisevar ;
		bfgs->gx[0] += (g1[i]-g2[i]) ;
		bfgs->gx[0] += 0.5*(temp-t1[i]*t2[i])*diag[i] ;
	}
	bfgs->gx[0] = - bfgs->gx[0] / 2.0 * settings->noisevar ;

	if ((NORMALIZED==settings->kernel||LINEAR==settings->kernel)&&1==settings->ardon)
		bfgs->gx[0] = 0 ;

	// for gx[1] --- \theta_b^1	
	for (i=0;i<settings->pairs->count;i++)
	{
		t1[i] = 0 ;
		for (j=0;j<settings->pairs->count;j++)
		{
			t1[i] += dsigma[i*settings->pairs->count+j]*(settings->alpha+j)->pair->epinvvar ;
		}
		t1[i] = t1[i]/(settings->alpha+i)->pair->epinvvar ;	
	}

	for (i=0;i<settings->pairs->count;i++)
	{
		bfgs->gx[1] -= (w1[i]-w2[i]) ;
		bfgs->gx[1] += 0.5*(t2[i])*(1.0-t1[i])*diag[i] ;
	}
	bfgs->gx[1] = bfgs->gx[1]/sigma ;

	// for gx[2] --- ln \Delta^r
	for (j=2;j<settings->pairs->classes;j++)
	{			
		// prepare D w / D delta_l
		for (ii=0;ii<settings->pairs->count;ii++)
		{
			if (j==(settings->alpha+ii)->pair->target)
				cache[ii] = -(g1[ii]+(w1[ii]-w2[ii])*w1[ii]) / settings->noisevar ;
			//else if (settings->pairs->classes==(settings->alpha+ii)->pair->target)
			//	cache[ii] = -(-g2[ii]-(w1[ii]-w2[ii])*w2[ii]) / settings->noisevar ;
			else if ((settings->alpha+ii)->pair->target>j)
				//cache[ii] = -(settings->alpha+ii)->pair->epinvvar ;
				cache[ii] = -( g1[ii]-g2[ii] + (w1[ii]-w2[ii])*(w1[ii]-w2[ii]) ) / settings->noisevar ;
			else
				cache[ii] = 0 ;
		}
		
		// prepare D f / D delta_l
		for (ii=0;ii<settings->pairs->count;ii++)
		{
			t1[ii] = 0 ;
			for (iii=0;iii<settings->pairs->count;iii++)
				t1[ii] += dsigma[ii*settings->pairs->count+iii] * cache[iii] ;
			t1[ii] = - t1[ii] / (settings->alpha+ii)->pair->epinvvar ;
		}

		// D f / D delta
		for (i=0;i<settings->pairs->count;i++)
		{
			if (j==(settings->alpha+i)->pair->target)
			{
				bfgs->gx[j] -= w1[i] ;
				temp = (w1[i]-2.0*(w1[i]-w2[i])*g1[i]
					-2.0*(w1[i]-w2[i])*(w1[i]-w2[i])*w1[i]-v1[i]
					-(g1[i]-g2[i])*w1[i]) / settings->noisevar ;
				//bfgs->gx[j] += 0.5*(temp-t2[i]*t1[i])*diag[i] ;
				bfgs->gx[j] += 0.5*(temp-t1[i]*t2[i])*diag[i] ;
			}
			/*else if ((settings->alpha+i)->pair->target==settings->pairs->classes)
			{
				bfgs->gx[j] += w2[i] ;
				temp =  (w1[i]-2.0*(w1[i]-w2[i])*g1[i]
					-2.0*(w1[i]-w2[i])*(w1[i]-w2[i])*w1[i]-v1[i]
					-(g1[i]-g2[i])*w1[i]) / settings->noisevar ;
				bfgs->gx[j] += 0.5*temp*diag[i] ;
			}*/
			else if ((settings->alpha+i)->pair->target>j)
			{
				bfgs->gx[j] -= (w1[i]-w2[i]) ;
				bfgs->gx[j] += 0.5*t2[i]*(1.0-t1[i])*diag[i] ;
			}
			else
				bfgs->gx[j] += 0.5*(-t1[i]*t2[i])*diag[i] ;
		}
		bfgs->gx[j] = bfgs->gx[j]*settings->intervals[j-1]/sigma ;
	}

	// for kappa_o
	/*bfgs->gx[bfgs->var_number-1] = 0 ;

		temp = 0 ;
		for (i=0;i<settings->pairs->count;i++)
		{
			for (j=0;j<i;j++)
				temp += 2.0*settings->invcov[i*settings->pairs->count+j]*(settings->alpha+i)->kernels[j] ;	
			temp += settings->invcov[i*settings->pairs->count+i]*(settings->alpha+i)->kernels[i] ;	
		}
		bfgs->gx[bfgs->var_number-1] = temp/2.0 ;
		// VC*VC*a'*dsigma*a/2
		for (i=0;i<settings->pairs->count;i++)
		{						
			for (j=0;j<i;j++)
				bfgs->gx[bfgs->var_number-1] -= (settings->alpha+i)->pair->weight*
				(settings->alpha+j)->pair->weight*(settings->alpha+i)->kernels[j] ;		
			bfgs->gx[bfgs->var_number-1] -= (settings->alpha+i)->pair->weight*
				(settings->alpha+i)->pair->weight*(settings->alpha+i)->kernels[i]/2.0 ;				
		}
		for (i=0;i<settings->pairs->count;i++)	
		{
			for (j=0;j<settings->pairs->count;j++)
			{
				dmat[i*settings->pairs->count+j] = 0 ;
				for (ii=0;ii<settings->pairs->count;ii++)
					dmat[i*settings->pairs->count+j] += settings->invcov[i*settings->pairs->count+ii]
						*Calculate_Covfun( settings->alpha+j, settings->alpha+ii, settings) ;			
			}
		}
		for (ii=0;ii<settings->pairs->count;ii++)
		{
			t1[ii] = 0 ;
			for (iii=0;iii<settings->pairs->count;iii++)
				t1[ii] += dmat[ii*settings->pairs->count+iii] * (settings->alpha+iii)->pair->weight ;
			t1[ii] = t1[ii] / (settings->alpha+ii)->pair->epinvvar ;			
		}
		for (ii=0;ii<settings->pairs->count;ii++)
		{
			bfgs->gx[bfgs->var_number-1] += diag[ii]*(-t2[ii])/sigma*t1[ii]/2.0 ;
		}
*/
	// for kernel parameters
	// calculate for each settings->kappa Gaussian parameter
	if (1 == settings->ardon)
	{
		for (ii=0;ii<settings->pairs->dimen;ii++)
		{			
			// calculate dsigmam for each settings->kappa parameter
			for (i=0;i<settings->pairs->count;i++)
			{
				alpha = settings->alpha + i ;	
				//if ( I_One == alpha->setname )
				{
					dsigma[i*settings->pairs->count+i] = 0 ;
					for (j=0;j<i;j++)
					{
						alphab = settings->alpha + j ;
						//if ( I_One == alphab->setname )
						{							
							if ( GAUSSIAN == settings->kernel )
							{
								dsigma[i*settings->pairs->count+j]=(alpha->kernels[j]-settings->kappa_m)*(-(alpha->pair->
									point[ii]-alphab->pair->point[ii])*(alpha->pair->point[ii]-alphab->pair->point[ii])/2.0/settings->pairs->dimen) ;
							}
							else if ( LINEAR == settings->kernel )
							{
								dsigma[i*settings->pairs->count+j]=(alpha->pair->point[ii]*alphab->pair->point[ii]) ;
							}
							else if ( NORMALIZED == settings->kernel )
                                                        {
								di=0; dj=0; dij=0;
								for (dindex=0;dindex<settings->pairs->dimen;dindex++)
								{
									di += settings->kappa[dindex]*alpha->pair->point[dindex]*alpha->pair->point[dindex] ;
									dj += settings->kappa[dindex]*alphab->pair->point[dindex]*alphab->pair->point[dindex] ; 
									dij += settings->kappa[dindex]*alpha->pair->point[dindex]*alphab->pair->point[dindex] ;
								}
								if (di==0||dj==0||dij==0)
									dsigma[i*settings->pairs->count+j] = 0 ;
								else
                                                                	dsigma[i*settings->pairs->count+j]=(alpha->kernels[j])*(alpha->pair->point[ii]*alphab->pair->point[ii]/(dij)-0.5*alpha->pair->point[ii]*alpha->pair->point[ii]/di-0.5*alphab->pair->point[ii]*alphab->pair->point[ii]/dj) ;
                                                        }

							/*
							else if ( POLYNOMIAL == settings->kernel ) 
							{
								cbe = 0 ; 
								for ( iii = 0 ; iii < settings->pairs->dimen; iii++)
									cbe += settings->kappa[iii]*alpha->pair->point[iii]*alphab->pair->point[iii] ;
								if (cbe > 0)
									dsigma[i*settings->pairs->count+j]=alpha->kernels[j]*settings->p*alpha->pair->point[ii]
									*alphab->pair->point[ii]/(cbe) ;
								else
									dsigma[i*settings->pairs->count+j]=0 ;
							}*/
							else if ( USERDEFINED == settings->kernel )
							{
								cbe = 0 ; // temp storage
								if (0==settings->pairs->featuretype[ii])
									cbe-=(alpha->pair->point[ii]-alphab->pair->point[ii]) 
									*(alpha->pair->point[ii]-alphab->pair->point[ii])/2.0 ;
								else if (alpha->pair->point[ii]!=alphab->pair->point[ii])
									cbe-=1.0/2.0 ;
								dsigma[i*settings->pairs->count+j] = (alpha->kernels[j]-settings->kappa_m) * cbe ;
							}
							else
								printf ("Error in kernel type.\n") ;
							dsigma[j*settings->pairs->count+i] = dsigma[i*settings->pairs->count+j] ;
						}											
					}
				}
			}
			//brmul(adjsgmm, dsigmam, settings->pairs->count, settings->pairs->count, settings->pairs->count, sigdsig) ;
			temp = 0 ;
			for (i=0;i<settings->pairs->count;i++)
			{
				for (j=0;j<i;j++)
					temp += 2.0*settings->invcov[i*settings->pairs->count+j]*dsigma[i*settings->pairs->count+j] ;	
				temp += settings->invcov[i*settings->pairs->count+i]*dsigma[i*settings->pairs->count+i] ;	
			}
			bfgs->gx[ii+settings->pairs->classes] = temp/2.0 ;
			// VC*VC*a'*dsigma*a/2
			for (i=0;i<settings->pairs->count;i++)
			{						
				for (j=i+1;j<settings->pairs->count;j++)
					bfgs->gx[ii+settings->pairs->classes] -= (settings->alpha+i)->pair->weight*
					(settings->alpha+j)->pair->weight*dsigma[i*settings->pairs->count+j] ;		
				bfgs->gx[ii+settings->pairs->classes] -= (settings->alpha+i)->pair->weight*
					(settings->alpha+i)->pair->weight*dsigma[i*settings->pairs->count+i]/2.0 ;				
			}
			
			for (i=0;i<settings->pairs->count;i++)	
			{
				for (j=0;j<settings->pairs->count;j++)
				{
					dmat[i*settings->pairs->count+j] = 0 ;
					for (iii=0;iii<settings->pairs->count;iii++)
						dmat[i*settings->pairs->count+j] += settings->invcov[i*settings->pairs->count+iii]
						*dsigma[j*settings->pairs->count+iii] ;			
				}
			}
			for (i=0;i<settings->pairs->count;i++)
			{
				t1[i] = 0 ;
				for (iii=0;iii<settings->pairs->count;iii++)
					t1[i] += dmat[i*settings->pairs->count+iii] * (w1[iii]-w2[iii])/sigma ;
				t1[i] = - t1[i] / (settings->alpha+i)->pair->epinvvar ;			
			}
			for (i=0;i<settings->pairs->count;i++)
			{
				bfgs->gx[ii+settings->pairs->classes] += diag[i]*(-t2[i])*t1[i]/2.0/sigma ;
			}
bfgs->gx[ii+settings->pairs->classes] += settings->regular*settings->kappa[ii] ;
bfgs->fx += 0.5*settings->regular*settings->kappa[ii]*settings->kappa[ii] ;
//bfgs->gx[ii+settings->pairs->classes] += settings->regular ;
//bfgs->fx += settings->regular*settings->kappa[ii] ;

			bfgs->gx[ii+settings->pairs->classes] = bfgs->gx[ii+settings->pairs->classes] * settings->kappa[ii] ;
		}
	}
	else
	{
		for (i=0;i<settings->pairs->count;i++)
		{
			alpha = settings->alpha + i ;
			dsigma[i*settings->pairs->count+i] = 0 ;			
			//if ( I_One == alpha->setname )
			{
				for (j=0;j<i;j++)
				{
					alphab = settings->alpha + j ;
					dsigma[i*settings->pairs->count+j] = 0 ;
					//if ( I_One == alphab->setname )
					{							
						if ( GAUSSIAN == settings->kernel )
						{
							cbe = 0 ; // temp storage
							for (ii=0;ii<settings->pairs->dimen;ii++)
								cbe-=(alpha->pair->point[ii]-alphab->pair->point[ii])
								*(alpha->pair->point[ii]-alphab->pair->point[ii]) ;
							dsigma[i*settings->pairs->count+j] = (alpha->kernels[j]-settings->kappa_m)*cbe/2.0/settings->pairs->dimen ;
						} 
						else if ( LINEAR == settings->kernel || NORMALIZED == settings->kernel )
						{
							dsigma[i*settings->pairs->count+j] = alpha->kernels[j]/settings->kappa_a ;
						}
						/*
						else if ( POLYNOMIAL == settings->kernel ) 
						{
							cbe = 0 ; 
							for (iii = 0 ; iii < settings->pairs->dimen; iii++)
								cbe += alpha->pair->point[iii]*alphab->pair->point[iii] ;
							if (cbe > 0)
								dsigma[i*settings->pairs->count+j] = alpha->kernels[j]*settings->p/settings->kappa_a ;
							else
								dsigma[i*settings->pairs->count+j] = 0 ;
						}*/
						else if ( USERDEFINED == settings->kernel )
						{
							cbe = 0 ; // temp storage
							for (ii=0;ii<settings->pairs->dimen;ii++)
							{
								if (0==settings->pairs->featuretype[ii])
									cbe-=(alpha->pair->point[ii]-alphab->pair->point[ii])
									*(alpha->pair->point[ii]-alphab->pair->point[ii]) ;
								else if (alpha->pair->point[ii]!=alphab->pair->point[ii])
									cbe-=1.0 ;
							}
							dsigma[i*settings->pairs->count+j] = (alpha->kernels[j]-settings->kappa_m) * cbe / 2.0 ;
						}
						else
							printf ("Error in kernel type.\n") ;
						dsigma[j*settings->pairs->count+i] = dsigma[i*settings->pairs->count+j] ;
					}
				}
			}
		}
		temp = 0 ;
		for (i=0;i<settings->pairs->count;i++)
		{
			for (j=0;j<i;j++)
				temp += 2.0*settings->invcov[i*settings->pairs->count+j]*dsigma[i*settings->pairs->count+j] ;	
			temp += settings->invcov[i*settings->pairs->count+i]*dsigma[i*settings->pairs->count+i] ;	
		}
		bfgs->gx[settings->pairs->classes] = temp/2.0 ;
		// VC*VC*a'*dsigma*a/2
		for (i=0;i<settings->pairs->count;i++)
		{						
			for (j=0;j<i;j++)
				bfgs->gx[settings->pairs->classes] -= (settings->alpha+i)->pair->weight*
				(settings->alpha+j)->pair->weight*dsigma[i*settings->pairs->count+j] ;		
			bfgs->gx[settings->pairs->classes] -= (settings->alpha+i)->pair->weight*
				(settings->alpha+i)->pair->weight*dsigma[i*settings->pairs->count+i]/2.0 ;				
		}
		for (i=0;i<settings->pairs->count;i++)	
		{
			for (j=0;j<settings->pairs->count;j++)
			{
				dmat[i*settings->pairs->count+j] = 0 ;
				for (ii=0;ii<settings->pairs->count;ii++)
					dmat[i*settings->pairs->count+j] += settings->invcov[i*settings->pairs->count+ii]
						*dsigma[j*settings->pairs->count+ii] ;			
			}
		}
		for (ii=0;ii<settings->pairs->count;ii++)
		{
			t1[ii] = 0 ;
			for (iii=0;iii<settings->pairs->count;iii++)
				t1[ii] += dmat[ii*settings->pairs->count+iii] * (settings->alpha+iii)->pair->weight ;
			t1[ii] = t1[ii] / (settings->alpha+ii)->pair->epinvvar ;			
		}
		for (ii=0;ii<settings->pairs->count;ii++)
		{
			bfgs->gx[settings->pairs->classes] += diag[ii]*(-t2[ii])/sigma*t1[ii]/2.0 ;
		}
		bfgs->gx[settings->pairs->classes] = bfgs->gx[settings->pairs->classes] * settings->kappa_a ;

	if (settings->kernel==LINEAR)
		bfgs->gx[settings->pairs->classes] = 0 ;
	}

 //       for (i=0;i<settings->number;i++)
 //       {
		//if (i<settings->pairs->classes)
		//	bfgs->gx[i]=0;
 //       }
//bfgs->gx[0] = 0 ;
	
#ifdef _ORDINAL_DEBUG
	//if (bfgs->x[0]>=100)
	//bfgs->gx[settings->number-1] = 0 ;
	printf("functional %f \n", bfgs->fx) ;
	for (i=0;i<settings->number;i++)
	{
		printf("gardient %f --- x %f\n", bfgs->gx[i], bfgs->x[i]) ;
	}
#endif 
	free(cache) ;
	free(diag) ;
	free(q1) ;
	free(q2) ;
	free(v1) ;
	free(v2) ;
	free(g1) ;
	free(g2) ;
	free(w2) ;
	free(w1) ;
	free(t2) ;
	free(t1) ;
	free(dsigma) ;	
	free(dmat) ;
	return 0 ;
}

int	Ordinal_LAPLACE_Training (Ordinal_Settings * settings) 
{
	unsigned int i ;

	if (NULL == settings)
		return 1 ;

	printf("Laplacian Approximation around MAP estimate ...\n") ;
	if ( 1 == bfgs_minfunc ( Ordinal_LAPLACE_Evaluate_FuncGrad, settings, settings->bfgs ) )
	{
		printf("fail to find local minimum.\n") ;
		return 1;
	}
	// save best funcgrad to settings 
	bfgs_ordinal_Settings( settings, settings->bfgs ) ;
	printf("Optimal Settings: \n") ;
	printf("Noise Variance: %.3f\n",settings->noisevar) ;
	printf("Kappa_O: %.3f\n",settings->kappa_o) ;
	printf("Thresholds:\n") ;
	for (i=1;i<settings->pairs->classes;i++)
		printf("%u ---  %.3f\n",i,settings->thresholds[i-1]) ;
	if (0==settings->ardon)		
		printf("Kernel: %f\n",settings->kappa_a) ;
	else if (settings->pairs->dimen<10) 
	{
		printf("Kernel:\n") ;
		for (i=0;i<settings->pairs->dimen;i++)
			printf("%u ---  %f:\n", i, settings->kappa[i]) ;
	}
	Ordinal_MAP_Training (settings) ;
	return 0 ;
}

