// header file --- ordinal.h 

// Gaussian Processes for Ordinal Regression

// Chu Wei (C) Copyright 2004


#ifdef  __cplusplus
extern "C" {
#endif

#ifndef _ORDINAL_REGRESSION
#define _ORDINAL_REGRESSION
	
#define MINNUM          (3)				// at least two
#define LENGTH          (307200)		// maximum value of line length in data file 
#define PI				(3.141592654)

//********CHOOSE ONE MODE ONLY**********//
//#define _ORDINAL_EP
//#define _ORDINAL_KFOLD
#define _ORDINAL_LP
//**************************************//

//#define _ORDINAL_DEBUG
#define _ORDINAL_LOO

typedef enum _Data_Type
{
	REGRESSION = 3 ,
	ORDINALREGRESSION = 5 ,
	MULTICLASS = 4 ,
	CLASSIFICATION = 1 , // indicate the noise model
	UNKNOWN = 0 ,

} Data_Type ;

typedef enum _Kernel_Name
{
	GAUSSIAN = 0 ,	
	POLYNOMIAL = 1 ,
	USERDEFINED = 3 ,
	EXPONENTIAL = 2 ,
	LINEAR = 4 ,
	NORMALIZED = 5 , //normalized linear

} Kernel_Name ;

typedef struct _Data_Node 
{
	long unsigned int index ;       // line number in data file loaded 
	double * point ;                // point to one input point
	int fold ;						// fold label 
	unsigned int target ;			// class label	
	unsigned int pred_label ;
	double pred_func ;
	double pred_prob ;              // predictive probability 
	double pred_var ;
	struct _Data_Node * next ;      // point to next node in the list

	double weight ;					// contribution for the predictor
	double epmean ;					// individual mean in EP
	double epinvvar ;					// individual 1/variance in EP
	double epamp ;					// individual amplitude in EP
	double postmean ;				// mean in posterior distribution

} Data_Node ;

typedef struct _Data_List 
{
	Data_Type datatype ;            // regression problem or classification
	int normalized_input ;          // point data_node normalized or not 
	int normalized_output ;         // target data_node normalized or not 
	
	unsigned long int count ;       // total number of samples 
	unsigned long int dimen ;            // dimension of input vector
	unsigned int classes ;			// number of ordinal variables 

	int kfold ;						// for k-fold only

	int * featuretype ;				// for userdefined Data_Type
	int * labels ;	
	int * labelnum ;

	//double * weights ;			// for linear models

	char * filename ;

	unsigned int i_ymax ;
	unsigned int i_ymin ;
	double mean ;                   // mean of output
	double deviation ;              // deviation of output
	double * x_mean ;               // mean of input
	double * x_devi ;               // standard deviation of input
	
	Data_Node * front ;             // point to first node in the list
	Data_Node * rear ;              // point to last node in the list

} Data_List ;


typedef struct _Alphas
{
	double alpha ;
	double beta ;
	double nu ;
	double loomean ;
	double loovar ;
	double z1 ;
	double z2 ;
	double hnew ; // new posterior mean
	double mnew ; // new individual mean
	double pnew ; // new individual variance
	double snew ; // new individual amplitude

	double * kernels ;
	double * postcov ;
	double * p_cache ;              // save Pi(x) for CG 
	
	Data_Node * pair ;              // point to the corresponding pair 

} Alphas ;

typedef struct _bfgs_FuncGrad
{
	unsigned int var_number ;
	double * x ;
	//double kaio ;	    // it is log(var) except C, 
	//Data_Type type ;
	double fx ;
	double * gx ;	
	double norm2gx ;    // true norm square |g(x)|^2
	double maskednorm ; // norm |g(x)|^2 of masked gradient 
	unsigned int iternum ;

} bfgs_FuncGrad ;

typedef struct _Ordinal_Settings
{
	Kernel_Name kernel ;
	unsigned int number ;
	unsigned int p  ;	
	unsigned int kfoldcv  ;
	int ardon ;
	int cacheall ;
	double kappa_o ;	
	double noisevar ;				// noise variance
	double kappa_a ;
	double * kappa ;				// kappa for ARD variables (d)
	double kappa_m ;
	double * thresholds ;			// ordinal threshold (r-1)
	double * intervals ; 			// threshold interval (r-2)
	
	double regular ;
	
	char * testfile ;               // the name of input data file 
	double * invcov ;
	
	bfgs_FuncGrad *	bfgs ;

	struct _Data_List * pairs ;     // this is a reference from def_Settings
	struct _Alphas * alpha ;        // Pointers to Alphas matrix 

} Ordinal_Settings ;

#define DEF_KFOLDCV			  (5)
#define DEF_NORMALIZEINPUT    (1)// 1 - YES TRUE ; 0 - NO FALSE
#define DEF_NORMALIZETARGET   (0)
#define DEF_CACHEALL		  (1)
#define DEF_REGULARIZATION    (0.1)
#define DEF_ARDON			  (1)
#define DEF_KAPPA			  (1)
#define DEF_NOISEVAR		  (0.5)
#define DEF_KERNEL			  (LINEAR)
#define DEF_KAPPA_M			  (0)
#define DEF_KAPPA_O			  (1)
#define DEF_P				  (2)
#define DEF_JITTER			  (0.001)
#define DEF_REGULAR			  (0.1)
#define EPS					  (0.000001)
#define TOL					  (0.001)

int bfgs_minfunc ( int (* Evaluate_FuncGrad)(void *, bfgs_FuncGrad *),	void * , bfgs_FuncGrad * ) ;

bfgs_FuncGrad * Create_bfgs_FuncGrad ( unsigned int number ) ;
void Clear_bfgs_FuncGrad ( bfgs_FuncGrad * funcgrad ) ;
int Compute_EP_Weights (Ordinal_Settings * settings) ;

double lerrf(double x) ;
double fromb_t1(int index, Ordinal_Settings * settings) ;
double fromb_t2(int index, Ordinal_Settings * settings) ;
double fromb_t3(int index, Ordinal_Settings * settings) ;
double fromb_t4(int index, Ordinal_Settings * settings) ;
double fromb_t5(int index, Ordinal_Settings * settings) ;

int Ordinal_EPEM_Evaluate_FuncGrad ( void * pointer, bfgs_FuncGrad * bfgs ) ;
int	Ordinal_EPEM_Training (Ordinal_Settings * settings) ;

int	Ordinal_LAPLACE_Training (Ordinal_Settings * settings) ;
int Ordinal_MAP_Training (Ordinal_Settings * settings) ;

double normal(double x) ;
int ordinal_Loadfile ( Data_List * pairs, char * inputfilename, int inputdim ) ;
int Is_Data_Empty ( Data_List * list ) ;
int Create_Data_List ( Data_List * list ) ;
int Clear_Data_List ( Data_List * list ) ;
int Add_Data_List ( Data_List * list, Data_Node * node ) ;
Data_Node * Create_Data_Node ( long unsigned int index, double * point, int y ) ;

int ordinal_bfgs_Settings ( Ordinal_Settings * settings, bfgs_FuncGrad * funcgrad ) ;
int bfgs_ordinal_Settings ( Ordinal_Settings * settings, bfgs_FuncGrad * funcgrad ) ;

Alphas * Create_Alphas (Ordinal_Settings * list) ;
int Clear_Alphas (Alphas * alpha, Data_List * list) ;

Ordinal_Settings * Create_Ordinal_Settings(Data_List * list) ;
int Clear_Ordinal_Settings( Ordinal_Settings * settings) ;

double Calculate_Covfun( struct _Alphas * ai, struct _Alphas * aj, Ordinal_Settings * settings ) ;
double Calc_Covfun( double * pi, double * pj, Ordinal_Settings * settings ) ;

int Ordinal_EP_Training (Ordinal_Settings *) ;
int Ordinal_Prediction (Ordinal_Settings * settings) ;

int Add_Label_Data_List ( Data_List * list, Data_Node * node ) ;
int Clear_Label_Data_List ( Data_List * list ) ;
//int Newton_Soft_Max (Data_List * list) ;

int Create_Kfold_List (Data_List * list, int KFOLDCV) ;
Data_Node ** Create_Fold_List (Data_List * list, Data_List * train, Data_List * test, int fold) ;
int Clear_Fold_List (Data_List * list, Data_List * train, Data_List * test, Data_Node **) ;
int Dumping_Ordinal_Settings (Ordinal_Settings * settings) ;

int Find_Out_Label (int * labels, int size, int inquiry) ;
int * Test_Soft_Max(Data_List * train, Data_List * test) ;
int Initial_Pcache (Alphas * alpha, Data_List * list) ;

double normpdf(double x) ;
double normcdf(double ca, double eb) ;
void indexx(unsigned long n, float arr[], unsigned long indx[]) ;

#endif

#ifdef  __cplusplus
}
#endif

