#include <stdio.h>
#include <stdlib.h>
#include <math.h>
#include <string.h>
#include <time.h>
#include "ordinal.h"
#include "neal_matrix.h"

double Retrieve_Posterior_Covariance (unsigned int i, unsigned int j, Ordinal_Settings * settings)
{
	if (NULL == settings)
		return 0 ;
	if (i>=settings->pairs->count||j>=settings->pairs->count)
	{
		printf("index is out of range.\n") ;
		return 0 ;
	}
	if (i >= j)
		return (settings->alpha+i)->postcov[j] ;
	else 
		return (settings->alpha+j)->postcov[i] ; 
}

double normal(double x)
{
	return exp(-0.5*x*x)/sqrt(2.0*PI) ;
}

double Ordinal_EP_Swipe(Ordinal_Settings * settings)
{
	double error = 0 ;
	unsigned int index ;
	double Aii, Zi ;
	double z1, z2, phi1, phi2, n1, n2, diff, rho, eta, dphi ;
	Alphas * alpha ;
	Data_Node * node ;
	double * ai ;
	unsigned int i , j, s=0 ;
	int * flag ;

	srand( (unsigned)time( NULL ) );
	//srand(2004) ;
	flag = (int *) calloc( settings->pairs->count,sizeof(int) ) ;
	//for (index=0;index<settings->pairs->count;index++)
	while (s<settings->pairs->count)
	{
		do 
		{
			index=(unsigned int)floor((double)rand()/(RAND_MAX+1.0)*settings->pairs->count) ;
		}
		while (flag[index]>0) ;
		flag[index] = 1 ;
		s += 1 ;
		// leave index out
		Aii = Retrieve_Posterior_Covariance (index, index, settings) ;
		alpha = settings->alpha+index ;
		node = alpha->pair ;
		
		alpha->loovar = Aii/(1-Aii*node->epinvvar) ;
		alpha->loomean = node->postmean + alpha->loovar*node->epinvvar*(node->postmean-node->epmean) ;

		// compute Z
		phi2 = 0 ;
		phi1 = 1 ;
		n1 = 0 ;
		n2 = 0 ;
		z1 = 0 ;
		z2 = 0 ;

		/*if (2 <= node->target)
		{
			z2 = (settings->thresholds[node->target-2] - alpha->loomean)/sqrt(alpha->loovar+settings->noisevar) ;
			phi2 = 0.5 + 0.5 * lerrf(z2/sqrt(2)) ;
			n2 = normal(z2) ;
		}
		if (settings->pairs->classes-1 >= node->target) 	
		{
			z1 = (settings->thresholds[node->target-1] - alpha->loomean)/sqrt(alpha->loovar+settings->noisevar) ;
			phi1 = 0.5 + 0.5 * lerrf(z1/sqrt(2)) ;
			n1 = normal(z1) ;
		}*/
		if (1 == node->target)
		{
			z1 = (settings->thresholds[node->target-1] - alpha->loomean)/sqrt(alpha->loovar+settings->noisevar) ;
			n1 = normpdf(z1) ;
			phi1 = 0.5 + normcdf(0,z1) ;
			dphi = phi1-phi2 ;

		}
		else if (settings->pairs->classes == node->target)
		{
			z2 = (settings->thresholds[node->target-2] - alpha->loomean)/sqrt(alpha->loovar+settings->noisevar) ;
			n2 = normpdf(z2) ;
			phi2 = 0.5 + normcdf(0,z2) ;
			dphi = phi1-phi2 ;
		}
		else
		{
			z1 = (settings->thresholds[node->target-1] - alpha->loomean)/sqrt(alpha->loovar+settings->noisevar) ;
			z2 = (settings->thresholds[node->target-2] - alpha->loomean)/sqrt(alpha->loovar+settings->noisevar) ;
			n1 = normpdf(z1) ;
			n2 = normpdf(z2) ;
			dphi = normcdf(z2,z1);
		}

		Zi = dphi ;
		if (Zi<EPS*EPS)
		{
			//approximate
			printf("get in approximate dphi\n") ;
			alpha->alpha = (z1*n1-z2*n2)/(n1-n2)/sqrt(alpha->loovar+settings->noisevar) ;
			alpha->beta = (z1*z1*n1-z2*z2*n2)/(n1-n2)/(alpha->loovar+settings->noisevar)/2.0 ;
		}
		else
		{
			alpha->alpha = -(n1-n2)/Zi/sqrt(alpha->loovar+settings->noisevar) ;
			alpha->beta = -(z1*n1-z2*n2)/Zi/(alpha->loovar+settings->noisevar)/2.0 ;
		}
#ifdef _ORDINAL_DEBUG
		//printf("%d :  Z=%f \n",index,Zi) ;		
		//printf("%d :  a=%f \n",index,alpha->alpha) ;	
		//printf("%d :  b=%f \n",index,alpha->beta) ;	
#endif
		alpha->nu = alpha->alpha*alpha->alpha-2.0*alpha->beta ;
		alpha->hnew = alpha->loomean + alpha->loovar * alpha->alpha ;
		alpha->pnew = alpha->nu/(1.0-alpha->loovar*alpha->nu) ;
		alpha->mnew = alpha->loomean + alpha->alpha/alpha->nu ;
		alpha->snew = Zi*sqrt(alpha->loovar*alpha->pnew+1.0)*exp(0.5*alpha->alpha*alpha->alpha/alpha->nu) ;

		// update posterior mean and cov
		diff = alpha->pnew - node->epinvvar ;
		if (fabs(diff)>(EPS)&&alpha->pnew>0)
		{
			rho = diff/(1+diff*Aii) ;
			eta = (alpha->alpha+node->epinvvar*(node->postmean-node->epmean))/(1.0-Aii*node->epinvvar) ;
			// prepare the ai vector ;
			ai = (double *) malloc( settings->pairs->count*sizeof(double) ) ;
			if (NULL==ai)
			{
				printf("fail to malloc ai\n") ;
				exit(1) ;
			}
			for (i=0;i<settings->pairs->count;i++)
				ai[i] = Retrieve_Posterior_Covariance (i, index, settings) ;
			// update posterior cov and mean
			for (i=0;i<settings->pairs->count;i++)
			{
				for (j=0;j<=i;j++)
				{
					(settings->alpha+i)->postcov[j]-=rho*ai[i]*ai[j] ;
				}
				(settings->alpha+i)->pair->postmean+=eta*ai[i] ;
			}
			free(ai) ;
			// update epmean epvar epamp
			error += (alpha->pnew-node->epinvvar)*(alpha->pnew-node->epinvvar) ;
			error += (alpha->mnew-node->epmean)*(alpha->mnew-node->epmean) ;
			error += (alpha->snew-node->epamp)*(alpha->snew-node->epamp) ;
			node->epinvvar = alpha->pnew ;
			node->epmean = alpha->mnew ;
			node->epamp = alpha->snew ;
#ifdef _ORDINAL_DEBUG
		//printf("%d :  epm=%f \n",index,node->epmean) ;		
		//printf("%d :  epp=%f \n",index,node->epinvvar) ;	
		//printf("%d :  eps=%f \n",index,node->epamp) ;	
#endif
		}
		else
		{	
			//error += EPS ;
#ifdef _ORDINAL_DEBUG
			printf("skip %d sample.\n", index+1) ;		
#endif		
		}
	}
	free(flag) ;
	return error ;
}

int Compute_EP_Weights (Ordinal_Settings * settings) 
{
	double * t1 ;
	double * t2 ;
	double temp ;
	unsigned int i , j ;

	if (NULL == settings)
		return 1 ;

	for (i=0;i<settings->pairs->count;i++)
	{
		if (0==(settings->alpha+i)->pair->epinvvar)
		{
			printf("%u sample has not been updated.\n",i+1) ;
			return 1 ;
		}
		else
			(settings->alpha+i)->pair->weight = (settings->alpha+i)->alpha ;
	}
	
	if (NULL != settings->invcov)
		free(settings->invcov) ;

	settings->invcov = (double*)malloc(settings->pairs->count*settings->pairs->count*sizeof(double)) ;
	t1 = (double*)malloc(settings->pairs->count*sizeof(double)) ;
	t2 = (double*)malloc(settings->pairs->count*sizeof(double)) ;
	for (i=0;i<settings->pairs->count;i++)
	{	
		for (j=0;j<i;j++)
		{
			settings->invcov[i*settings->pairs->count+j] = (settings->alpha+i)->kernels[j] ;
			settings->invcov[j*settings->pairs->count+i] = settings->invcov[i*settings->pairs->count+j] ;
		}
		settings->invcov[i*settings->pairs->count+i] = (settings->alpha+i)->kernels[i] ;
		settings->invcov[i*settings->pairs->count+i] += 1.0/(settings->alpha+i)->pair->epinvvar ;
	}
	if (0 == cholesky(settings->invcov,settings->pairs->count,&temp))
		printf("Fatal Error: Sigma+1/Pi is not positive definite.\n") ;
	if (0 == inverse_from_cholesky(settings->invcov, t1, t2, settings->pairs->count) )		
		printf("Fatal Error: Sigma+1/Pi is not positive definite.\n") ;
	
	for (i=0;i<settings->pairs->count;i++)
	{
		(settings->alpha+i)->pair->weight = 0 ;
		for (j=0;j<settings->pairs->count;j++)
			(settings->alpha+i)->pair->weight += settings->invcov[i*settings->pairs->count+j]*(settings->alpha+j)->pair->epmean ;
		// check the equilibrium
		if (fabs((settings->alpha+i)->alpha-(settings->alpha+i)->pair->weight)>sqrt(EPS))
			printf("%u sample is not at equilibrium %f---%f.\n",i+1,(settings->alpha+i)->alpha,(settings->alpha+i)->pair->weight) ;
	}
	free(t1);
	free(t2);
	return 0 ;	
}

int Ordinal_EP_Training (Ordinal_Settings * settings) 
{
	double error ;
	int index = 0 ;
	//unsigned int i, j ;

	if (NULL == settings)
		return 0 ;

	// may refresh ALPHAS here 
	if (NULL != settings->alpha)
		Clear_Alphas(settings->alpha, settings->pairs) ;
	settings->alpha = Create_Alphas (settings) ;
	if (NULL == settings->alpha)
	{
		printf("fail to create ALPHAs.\n") ;
		exit(1) ;
	}
	do
	{
		error = Ordinal_EP_Swipe(settings) ; // clear sign
#ifdef _ORDINAL_DEBUG
		printf("Loop %d --- %f\n", index, error) ;
#endif
		index += 1 ;
	}
	while (error/(double)settings->pairs->count>EPS) ;

	if (1 == Compute_EP_Weights (settings))
		return 1 ;

#ifdef _ORDINAL_DEBUG
	/*for (index=0;index<(int)settings->pairs->count;index++)

		printf("%u --- post m %.3f - m %.3f -p %.3f -s %.3f -a %.3f -ai %.3f -vi %.3f\n", 
		(settings->alpha+index)->pair->target, 
		(settings->alpha+index)->pair->postmean,
		(settings->alpha+index)->pair->epmean,
		(settings->alpha+index)->pair->epinvvar,
		(settings->alpha+index)->pair->epamp, 
		(settings->alpha+index)->pair->weight,
		(settings->alpha+index)->alpha,
		(settings->alpha+index)->nu) ;*/
#endif
	return 0 ;
}


