#include <stdio.h>
#include <stdlib.h>
#include <math.h>
#include <string.h>
#include <time.h>
#include "ordinal.h"
#include "neal_matrix.h"

#define UPPERFUNC (10.0)

int Compute_MAP_Weights (Ordinal_Settings * settings) 
{	
	unsigned int i ;
	double z1, z2, phi1, phi2, n1, n2 ;
	double sigma, dphi ;

	if (NULL == settings)
	{
		printf("Input pointer is NULL.\n") ;
		return 1 ;
	}
	if (settings->noisevar<=0)
		return 1 ;

	sigma = sqrt(settings->noisevar) ;
	
	// prepare w_i
	for (i=0;i<settings->pairs->count;i++)
	{
		// update alpha_old
		(settings->alpha+i)->alpha = (settings->alpha+i)->pair->epamp ;
		(settings->alpha+i)->pair->epmean = (settings->alpha+i)->pair->postmean ;
		
		//if (fabs((settings->alpha+i)->pair->postmean)>UPPERFUNC)
		//	(settings->alpha+i)->pair->postmean = UPPERFUNC ;
			
		n1 = 0 ;
		n2 = 0 ;
		z1 = 0 ;
		z2 = 0 ;
		phi1 = 1 ;
		phi2 = 0 ;

		/*if ( 2 <= (settings->alpha+i)->pair->target )
		{
			z2 = (settings->thresholds[(settings->alpha+i)->pair->target-2]-(settings->alpha+i)->pair->epmean)/sigma ;
			phi2 = 0.5 + 0.5 * lerrf(z2/sqrt(2)) ;
			n2 = normal(z2) ;
		}
		if ( settings->pairs->classes-1 >= (settings->alpha+i)->pair->target )
		{
			z1 = (settings->thresholds[(settings->alpha+i)->pair->target-1]-(settings->alpha+i)->pair->epmean)/sigma ;
			phi1 = 0.5 + 0.5 * lerrf(z1/sqrt(2)) ;
			n1 = normal(z1) ;
		}*/
		if (1 == (settings->alpha+i)->pair->target)
		{
			z1 = (settings->thresholds[(settings->alpha+i)->pair->target-1]-(settings->alpha+i)->pair->epmean)/sigma ;
			n1 = normpdf(z1) ;
			phi1 = 0.5 + normcdf(0,z1) ;
			dphi = phi1-phi2 ;

		}
		else if (settings->pairs->classes == (settings->alpha+i)->pair->target)
		{
			z2 = (settings->thresholds[(settings->alpha+i)->pair->target-2]-(settings->alpha+i)->pair->epmean)/sigma ;
			n2 = normpdf(z2) ;
			phi2 = 0.5 + normcdf(0,z2) ;
			dphi = phi1-phi2 ;
		}
		else
		{
			z1 = (settings->thresholds[(settings->alpha+i)->pair->target-1]-(settings->alpha+i)->pair->epmean)/sigma ;
			z2 = (settings->thresholds[(settings->alpha+i)->pair->target-2]-(settings->alpha+i)->pair->epmean)/sigma ;			
			n1 = normpdf(z1) ;
			n2 = normpdf(z2) ;
			dphi = normcdf(z2,z1);
		}

		if (dphi<EPS)
		{
			if (1 == (settings->alpha+i)->pair->target)
			{
				(settings->alpha+i)->pair->weight = -z1/sigma ;
				(settings->alpha+i)->pair->epinvvar = 1/sigma/sigma ;
			}
			else if (settings->pairs->classes == (settings->alpha+i)->pair->target)
			{
				(settings->alpha+i)->pair->weight = z2/sigma ;
				(settings->alpha+i)->pair->epinvvar = 1/sigma/sigma ;
			}
			else
			{
				if ((n1-n2)>=0)
				{
					(settings->alpha+i)->pair->weight = -(z1*exp(-0.5*z1*z1+0.5*z2*z2)-z2)/(exp(-0.5*z1*z1+0.5*z2*z2)-1.0)/sigma ;
					(settings->alpha+i)->pair->epinvvar = 1/sigma/sigma ;
					//(settings->alpha+i)->pair->epinvvar = 1/sigma/sigma 
					//	+ (settings->alpha+i)->pair->weight*(settings->alpha+i)->pair->weight
					//	- (z1*z1*exp(-0.5*z1*z1+0.5*z2*z2)-z2*z2)/(exp(-0.5*z1*z1+0.5*z2*z2)-1.0)/sigma/sigma ;
				}
				else
				{
					//printf("final weight: n1 < n2 ;\n") ;
					(settings->alpha+i)->pair->weight = 0 ;
					(settings->alpha+i)->pair->epmean = 0 ;
					(settings->alpha+i)->pair->postmean = 0 ;
					(settings->alpha+i)->pair->epinvvar = 1/sigma/sigma ;
				}
			}
#ifdef _ORDINAL_DEBUG
			//printf("%u approximation, w=%f, p=%f\n",i,(settings->alpha+i)->pair->weight,(settings->alpha+i)->pair->epinvvar) ;
#endif
		}
		else
		{
			(settings->alpha+i)->pair->weight = (n1-n2)/(dphi)/sigma ;
			(settings->alpha+i)->pair->epinvvar = (settings->alpha+i)->pair->weight*(settings->alpha+i)->pair->weight 
				+ (z1*n1-z2*n2)/(dphi)/sigma/sigma ;
		}

		(settings->alpha+i)->pair->weight = -(settings->alpha+i)->pair->weight ;

		if ((settings->alpha+i)->pair->epinvvar>1.0/sigma/sigma)
		{
			//printf("final weight: %u, %f\n", i, (settings->alpha+i)->pair->epinvvar) ;
			(settings->alpha+i)->pair->epinvvar = 1.0 - EPS*EPS ;
		}
		
		if ((settings->alpha+i)->pair->epinvvar<0)
		{
			//printf("final weight: %u, %f\n", i, (settings->alpha+i)->pair->epinvvar) ;
			(settings->alpha+i)->pair->epinvvar = EPS*EPS ;
		}
	}
	return 0 ;
}

double Ordinal_MAP_Update(Ordinal_Settings * settings)
{
	double error = 0 ;
	double sigma ;
	double temp ;
	double * t1 ;
	double * t2 ;
	double * f ;	
	unsigned int i, j ;
	double z1, z2, phi1, phi2, n1, n2, dphi ;
	double func = 0 ;
	double step = 1.0 ;

	if (NULL == settings)
	{
		printf("Input pointer is NULL.\n") ;
		return  0 ;
	}
	if (settings->noisevar<=0)
		return 0 ;

	sigma = sqrt(settings->noisevar) ;
	if (NULL != settings->invcov) 
		free (settings->invcov) ;

	settings->invcov = (double *)malloc(settings->pairs->count*settings->pairs->count*sizeof(double)) ;
	f = (double *)calloc(settings->pairs->count,sizeof(double)) ;

	// prepare w_i
	for (i=0;i<settings->pairs->count;i++)
	{
		// update alpha_old
		(settings->alpha+i)->alpha = (settings->alpha+i)->pair->epamp ;
		(settings->alpha+i)->pair->epmean = (settings->alpha+i)->pair->postmean ;
	
		n1 = 0 ;
		n2 = 0 ;
		z1 = 0 ;
		z2 = 0 ;
		phi1 = 1 ;
		phi2 = 0 ;

		if (1 == (settings->alpha+i)->pair->target)
		{
			z1 = (settings->thresholds[(settings->alpha+i)->pair->target-1]-(settings->alpha+i)->pair->epmean)/sigma ;
			n1 = normpdf(z1) ;
			phi1 = 0.5 + normcdf(0,z1) ;
			dphi = phi1-phi2 ;

		}
		else if (settings->pairs->classes == (settings->alpha+i)->pair->target)
		{
			z2 = (settings->thresholds[(settings->alpha+i)->pair->target-2]-(settings->alpha+i)->pair->epmean)/sigma ;
			n2 = normpdf(z2) ;
			phi2 = 0.5 + normcdf(0,z2) ;
			dphi = phi1-phi2 ;
		}
		else
		{
			z1 = (settings->thresholds[(settings->alpha+i)->pair->target-1]-(settings->alpha+i)->pair->epmean)/sigma ;
			z2 = (settings->thresholds[(settings->alpha+i)->pair->target-2]-(settings->alpha+i)->pair->epmean)/sigma ;			
			n1 = normpdf(z1) ;
			n2 = normpdf(z2) ;
			dphi = normcdf(z2,z1);
		}

		if (dphi<EPS)
		{
			if (n1-n2<EPS)
			{
				func -= log(EPS) ;
				step = 0.01 ;		
			}
			else
			{
				func -= log(EPS) ;
				step = 0.1 ;
			}

			if (1 == (settings->alpha+i)->pair->target)
			{
				(settings->alpha+i)->pair->weight = -z1/sigma ;
				(settings->alpha+i)->pair->epinvvar = 1/sigma/sigma ;
			}
			else if (settings->pairs->classes == (settings->alpha+i)->pair->target)
			{
				(settings->alpha+i)->pair->weight = z2/sigma ;
				(settings->alpha+i)->pair->epinvvar = 1/sigma/sigma ;
			}
			else
			{
				if ((n1-n2)>=0)
				{
					(settings->alpha+i)->pair->weight = -(z1*exp(-0.5*z1*z1+0.5*z2*z2)-z2)/(exp(-0.5*z1*z1+0.5*z2*z2)-1.0)/sigma ;
					(settings->alpha+i)->pair->epinvvar = 1/sigma/sigma ;
					//(settings->alpha+i)->pair->epinvvar = 1/sigma/sigma 
					//	+ (settings->alpha+i)->pair->weight*(settings->alpha+i)->pair->weight
					//	- (z1*z1*exp(-0.5*z1*z1+0.5*z2*z2)-z2*z2)/(exp(-0.5*z1*z1+0.5*z2*z2)-1.0)/sigma/sigma ;
				}
				else
				{
					printf("n1 < n2 ;\n") ;
					(settings->alpha+i)->pair->weight = 0 ;
					(settings->alpha+i)->pair->epmean = 0 ;
					(settings->alpha+i)->pair->postmean = 0 ;
					(settings->alpha+i)->pair->epinvvar = 1/sigma/sigma ;
				}
			}
#ifdef _ORDINAL_DEBUG
			//printf("%u approximation, w=%f, p=%f\n",i,(settings->alpha+i)->pair->weight,(settings->alpha+i)->pair->epinvvar) ;
#endif
		}
		else
		{
			(settings->alpha+i)->pair->weight = (n1-n2)/(dphi)/sigma ;
			(settings->alpha+i)->pair->epinvvar = (settings->alpha+i)->pair->weight*(settings->alpha+i)->pair->weight 
				+ (z1*n1-z2*n2)/(dphi)/sigma/sigma ;
			// update func 			
			func -= log(dphi) ;
		}

		(settings->alpha+i)->pair->weight = -(settings->alpha+i)->pair->weight ;

		if ((settings->alpha+i)->pair->epinvvar>1.0/sigma/sigma)
		{
			//printf("error : %u, %f\n", i, (settings->alpha+i)->pair->epinvvar) ;
			(settings->alpha+i)->pair->epinvvar = 1.0 - EPS*EPS ;
		}
		
		if ((settings->alpha+i)->pair->epinvvar<0)
		{
			//printf("error : %u, %f\n", i, (settings->alpha+i)->pair->epinvvar) ;
			(settings->alpha+i)->pair->epinvvar = EPS*EPS ;
		}
	}

	for (i=0;i<settings->pairs->count;i++)
	{
		for (j=0;j<i;j++)
		{
			settings->invcov[i*settings->pairs->count+j] = 
				Calculate_Covfun(settings->alpha+i,settings->alpha+j,settings) ;
			settings->invcov[j*settings->pairs->count+i] = 
				settings->invcov[i*settings->pairs->count+j] ;
			f[i] -= Calculate_Covfun(settings->alpha+i,settings->alpha+j,settings) * (settings->alpha+j)->pair->weight ;
			f[j] -= Calculate_Covfun(settings->alpha+i,settings->alpha+j,settings) * (settings->alpha+i)->pair->weight ;
		}
		f[i] -= Calculate_Covfun(settings->alpha+i,settings->alpha+i,settings) * (settings->alpha+i)->pair->weight ;
		settings->invcov[i*settings->pairs->count+i] = Calculate_Covfun(settings->alpha+i,settings->alpha+i,settings) ;
		settings->invcov[i*settings->pairs->count+i] += 1.0/(settings->alpha+i)->pair->epinvvar ;
		f[i] += (settings->alpha+i)->pair->epmean ;
#ifdef _ORDINAL_DEBUG
		//printf("w(%u) = %f ---- p = %f --- f(%u) = %f\n",i,(settings->alpha+i)->pair->weight,(settings->alpha+i)->pair->epinvvar,i,f[i]) ;
#endif
	}
	t1 = (double *)malloc(settings->pairs->count*sizeof(double)) ;
	t2 = (double *)malloc(settings->pairs->count*sizeof(double)) ;
	if (0 == cholesky(settings->invcov,settings->pairs->count,&temp))
		printf("Fatal Error: Sigma+1/Pi is not positive definite.\n") ;
	if (0 == inverse_from_cholesky(settings->invcov, t1, t2, settings->pairs->count) )		
		printf("Fatal Error: Sigma+1/Pi is not positive definite.\n") ;

	for (i=0;i<settings->pairs->count;i++)
	{
		t1[i] = 0 ;
		for (j=0;j<i;j++)
		{
			t1[i] -= settings->invcov[i*settings->pairs->count+j]*f[j] ;			
			t1[j] -= settings->invcov[j*settings->pairs->count+i]*f[i] ;	
		}
		t1[i] -= settings->invcov[i*settings->pairs->count+i]*f[i] ;	
	}
	for (i=0;i<settings->pairs->count;i++)
	{	
		t1[i] = t1[i]/(settings->alpha+i)->pair->epinvvar ;
		(settings->alpha+i)->pair->postmean += step*t1[i] ; 	
		if (fabs((settings->alpha+i)->pair->postmean)>UPPERFUNC)
		{
#ifdef _ORDINAL_DEBUG
			printf("%u - new %f --- old %f ---- w %f \n",i, (settings->alpha+i)->pair->postmean,(settings->alpha+i)->pair->epmean,(settings->alpha+i)->pair->weight) ;
#endif
			//step = (UPPERFUNC-fabs((settings->alpha+i)->pair->epmean))/fabs(t1[i]) ;
			(settings->alpha+i)->pair->postmean = UPPERFUNC ;
			
		}
		if (error<fabs(t1[i]))		
			error = fabs(t1[i]) ;	
	}
	free (t2) ;
	free (t1) ;	
#ifdef _ORDINAL_DEBUG
	if (NULL != settings->invcov) 
		free (settings->invcov) ;
	settings->invcov = (double *)malloc(settings->pairs->count*settings->pairs->count*sizeof(double)) ;
	for (i=0;i<settings->pairs->count;i++)
	{
		for (j=0;j<i;j++)
		{
			settings->invcov[i*settings->pairs->count+j] = (settings->alpha+i)->kernels[j] ; 
			settings->invcov[j*settings->pairs->count+i] = (settings->alpha+i)->kernels[j] ;
		}
		settings->invcov[i*settings->pairs->count+i] = (settings->alpha+i)->kernels[i] ;
	}
	t1 = (double *)malloc(settings->pairs->count*sizeof(double)) ;
	t2 = (double *)malloc(settings->pairs->count*sizeof(double)) ;
	if (0 == cholesky(settings->invcov,settings->pairs->count,&temp))
		printf("Fatal Error: Sigma+1/Pi is not positive definite.\n") ;
	if (0 == inverse_from_cholesky(settings->invcov, t1, t2, settings->pairs->count) )		
		printf("Fatal Error: Sigma+1/Pi is not positive definite.\n") ;
	for (i=0;i<settings->pairs->count;i++)
	{
		for (j=0;j<i;j++)
		{
			func += settings->invcov[i*settings->pairs->count+j] * (settings->alpha+j)->pair->postmean * (settings->alpha+i)->pair->postmean ;
		}
		func += 0.5 * settings->invcov[i*settings->pairs->count+i] * (settings->alpha+i)->pair->postmean * (settings->alpha+i)->pair->postmean ;
	}
	printf("FUNCTIONAL : %f \n", func) ;
	free (t2) ;
	free (t1) ;	
#endif 

	free(f) ;
	return error ;
}

int Ordinal_MAP_Training (Ordinal_Settings * settings) 
{
	double error ;
	int index = 0 ;
	unsigned int i ;

	if (NULL == settings)
		return 0 ;

	// may refresh ALPHAS here 
	if (NULL != settings->alpha)
		Clear_Alphas(settings->alpha, settings->pairs) ;
	settings->alpha = Create_Alphas (settings) ;
	if (NULL == settings->alpha)
	{
		printf("fail to create ALPHAs.\n") ;
		exit(1) ;
	}
#ifdef _ORDINAL_DEBUG
	printf("Current Settings: \n") ;
	printf("Noise Variance: %.3f\n",settings->noisevar) ;
	printf("Kappa_O: %.3f\n",settings->kappa_o) ;
	printf("Thresholds:\n") ;
	for (i=1;i<settings->pairs->classes;i++)
		printf("%u ---  %.3f\n",i,settings->thresholds[i-1]) ;
	if (0==settings->ardon)		
		printf("Kernel: %.3f\n",settings->kappa_a) ;
	else
	{
		printf("Kernel:\n") ;
		for (i=0;i<settings->pairs->dimen;i++)
			printf("%u ---  %.3f:\n", i, settings->kappa[i]) ;
	}
#endif

	// intialize function
	for (i=0;i<settings->pairs->count;i++)
	{
		if (1 == (settings->alpha+i)->pair->target)
			(settings->alpha+i)->pair->postmean = settings->thresholds[(settings->alpha+i)->pair->target-1] ;
		else
			(settings->alpha+i)->pair->postmean = settings->thresholds[(settings->alpha+i)->pair->target-2] ;
	}
	do
	{
		error = Ordinal_MAP_Update(settings) ; // clear sign
		index += 1 ;
#ifdef _ORDINAL_DEBUG
		printf("Loop %d --- %f\n", index, error) ;
#endif
	}
	while (error>TOL&&index<250) ;

	if (1 == Compute_MAP_Weights (settings))
		return 1 ;

#ifdef _ORDINAL_DEBUG
	/*for (index=0;index<(int)settings->pairs->count;index++)
		printf("%u --- post m %.3f - weight %.3f -alpha %.3f \n", 
		(settings->alpha+index)->pair->target, 
		(settings->alpha+index)->pair->postmean,
		(settings->alpha+index)->pair->weight,
		(settings->alpha+index)->alpha) ;*/
#endif
	return 0 ;
}



